package de.dfki.sds.hephaistos.deeplink;

import java.net.URI;
import java.net.URL;
import java.net.URLConnection;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;
import java.util.StringJoiner;
import org.apache.commons.io.IOUtils;
import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.rdf.model.Property;
import org.apache.jena.rdf.model.RDFNode;
import org.apache.jena.rdf.model.Resource;
import org.apache.jena.rdf.model.ResourceFactory;

/**
 *
 * @author Markus Schr&ouml;der
 */
public class DeepLink {

    public static final String SEPARATOR = "@";
    public static final String SEPARATOR_PARAM = ",";
    public static final String DONTCARE = "*";

    private DeepLinker deeplinker;
    private LinkedList<DLResource> resources;
    private LinkedList<String> fragids;
    private LinkedList<String> methods;
    private LinkedList<Object[]> params;

    public DeepLink(DeepLinker deeplinker) {
        this.deeplinker = deeplinker;
        resources = new LinkedList<>();
        fragids = new LinkedList<>();
        methods = new LinkedList<>();
        params = new LinkedList<>();
    }

    public LinkedList<DLResource> getResources() {
        return resources;
    }

    public LinkedList<String> getFragids() {
        return fragids;
    }

    public LinkedList<String> getMethods() {
        return methods;
    }

    public String getMethod(int last) {
        return methods.get(methods.size() - 1 + last);
    }

    public LinkedList<Object[]> getParams() {
        return params;
    }

    public Object[] getParams(int last) {
        return params.get(params.size() - 1 + last);
    }

    public DLResource last() {
        return resources.getLast();
    }

    public <T extends DLResource> T last(Class<T> type) {
        return (T) resources.getLast();
    }

    public DLResource last(int index) {
        int targetIndex = resources.size() - 1 + index;
        if (targetIndex < 0) {
            return null;
        }
        return resources.get(targetIndex);
    }

    public DeepLink copy() {
        return copy(0);
    }

    public DeepLink copy(int last) {
        DeepLink ppr = new DeepLink(deeplinker);
        for (int i = 0; i < resources.size() + last; i++) {
            ppr.resources.add(resources.get(i));
        }
        for (int i = 0; i < fragids.size() + last; i++) {
            ppr.fragids.add(fragids.get(i));
            ppr.params.add(params.get(i));
            ppr.methods.add(methods.get(i));
        }
        return ppr;
    }

    /**
     * ends with '/'.
     *
     * @return
     */
    public String getPath() {
        return getPath(0);
    }

    //use prune = -1 to prune the path
    public String getPath(int prune) {
        StringBuilder sb = new StringBuilder();
        sb.append("/");
        int index = 0;
        for (String part : fragids) { //fragids are already correct encoded
            sb.append(part);
            sb.append("/");

            index++;

            if (index >= fragids.size() + prune) {
                break;
            }
        }
        String path = sb.toString();
        //remove ending /
        if (path.endsWith("/")) {
            return path.substring(0, path.length() - 1);
        }
        return path;
    }

    public DeepLink reinterpret() {
        try {
            return deeplinker.processPath(getPath());
        } catch (Exception ex) {
            throw new RuntimeException(ex);
        }
    }

    public String getHyperlink() {
        return deeplinker.getHost() + getPath();
    }

    public String getParamString(Object... params) {
        StringJoiner paramJoiner = new StringJoiner(SEPARATOR_PARAM);
        for (Object param : params) {
            paramJoiner.add(DeepLinker.encode2(param.toString()));
        }
        return paramJoiner.toString();
    }

    public String getHyperlink(String method, Object... params) {
        return deeplinker.getHost() + getPath() + "/" + getFragId(method, params);
    }

    public String getFragId(String method, Object... params) {
        //jörn suggestion: child is default
        if (method.equals("child")) {
            return getParamString(params);
        }
        return method + SEPARATOR + getParamString(params);
    }

    public DeepLink append(String method, Object... params) {
        DeepLink copy = copy();
        copy.fragids.add(getFragId(method, params));
        copy.methods.add(method);
        copy.params.add(params);
        return copy;
    }

    public DeepLink replace(String method, Object... params) {
        return copy(-1).append(method, params);
    }

    public DeepLink next(boolean replace, String method, Object... params) {
        DeepLink dl;
        if (method == null) {
            dl = this;
        } else if (replace) {
            dl = this.replace(method, params);
        } else {
            dl = this.append(method, params);
        }
        return dl;
    }

    public String getHyperlinkContent(String method, Object... params) {
        String contentSegment = "content/";

        //jörn suggestion: child is default
        if (method.equals("child")) {
            return deeplinker.getHost() + getPath() + "/" + contentSegment + getParamString(params);
        }

        return deeplinker.getHost() + getPath() + "/" + contentSegment + method + SEPARATOR + getParamString(params);
    }

    public String getHyperlink(boolean replace, String method, Object... params) {
        if (replace) {
            return getHyperlinkReplace(-1, method, params);
        } else {
            return getHyperlink(method, params);
        }
    }

    public String getHyperlink(boolean[] replaces, String method, Object... params) {
        int sum = 0;
        for (boolean b : replaces) {
            if (b) {
                sum -= 1;
            }
        }

        return getHyperlinkReplace(sum, method, params);
    }

    public String getHyperlinkReplace(int shorten, String method, Object... params) {
        //jörn suggestion: child is default
        if (method.equals("child")) {
            return deeplinker.getHost() + getPath(shorten) + "/" + getParamString(params);
        }

        return deeplinker.getHost() + getPath(shorten) + "/" + method + SEPARATOR + getParamString(params);
    }

//    private void annotationTableHtmlContent(JSONObject annotation, StringBuilder html) {
//        for (JSONArray row : new JSON(annotation.getJSONArray("annoTable")).toList(JSONArray.class)) {
//            //data-ng-repeat=\"row in annotation.annoTable\"
//            html.append("<tr>"); //style=\"z-index: 1; position:relative\"
//
//            for (JSONObject col : new JSON(row).toList(JSONObject.class)) {
//                //data-ng-repeat=\"col in row track by $index\"
//                html.append("<td>");
//
//                //<a href=\"\" title=\"\" class=\"annotation-anchor\" name=\"{{col.binding.annotation}}\" ng-click=\"clickAnnotationCol(col)\" ng-mouseenter=\"mouseenterAnnotationCol(col)\" ng-mouseleave=\"mouseleaveAnnotationCol()\">{{col.c}}</a>
//                html.append("<a ");
//                html.append("onContextMenu=\"return false;\" ");
//                html.append("title=\"\" ");
//                html.append("class=\"annotation-anchor\" ");
//
//                if (col.has("binding")) {
//                    String annoURI = col.getJSONObject("binding").getString("annotation");
//
//                    html.append("onmouseup=\"if(event.which === 3){ /*if (confirm('Delete annotation?')){*/");
//                    html.append("annotationRemove($(event.target).closest('table'), '" + annoURI +"');");
//                    html.append("/*}*/}\" ");
//
//                    html.append("name=\"" + annoURI + "\" ");
//                    html.append("href=\"" + col.getJSONObject("binding").getString("target") + "\" ");
//                }
//
//                html.append(">");
//                html.append(col.getString("c"));
//                html.append("</a>");
//
//                html.append("</td>");
//            }
//
//            html.append("</tr>");
//        }
//
//        if (annotation.getJSONArray("annoTable").length() > 0) {
//            html.append("<tr style=\"line-height: 0px;\">");
//
//            for (int i = 0; i < annotation.getString("fullText").length(); i++) {
//                html.append("<td>");
//                html.append("<div style=\"user-select: none; z-index: -1; position:relative\">&nbsp;</div>");
//                html.append("</td>");
//            }
//
//            html.append("</tr>");
//        }
//    }
    
    
    private String getLiteral(Resource subject, Property predicate, Model model) {
        List<RDFNode> l = model.listObjectsOfProperty(subject, predicate).toList();
        if (l.isEmpty()) {
            return null;
        }
        return l.get(0).asLiteral().getLexicalForm();
    }

    /**
     * host ist without ending /.
     *
     * @return
     */
    public String getHost() {
        return deeplinker.getHost();
    }

    @Override
    public String toString() {
        return getHyperlink();
    }

    public DeepLinker deeplinker() {
        return deeplinker;
    }

    public Model downloadRDF() {
        //we can use deeplinkers rdf con.neg. to get initial data about it (inital model)
        //all about the file
        Model m = ModelFactory.createDefaultModel();
        m.setNsPrefixes(deeplinker().getPrefixMapping());
        try {
            m.read(getHyperlink(), "TTL");
        } catch (Exception e) {
            //try convert to@rdf
            try {
                m.read(append("to", "rdf").getHyperlink(), "TTL");
            } catch (Exception e2) {
                //ignore if not supported; model is then empty
            }
        }
        return m;
    }

    public byte[] downloadBinary() {
        return download("application/octet-stream");
    }

    public byte[] download(String accept) {
        try {
            URLConnection con = new URL(getHyperlink()).openConnection();
            con.addRequestProperty("Accept", accept);
            return IOUtils.toByteArray(con);
        } catch (Exception ex) {
            throw new RuntimeException(ex);
        }
    }

    public org.apache.jena.rdf.model.Resource asResource() {
        return ResourceFactory.createResource(getHyperlink());
    }

    public org.apache.jena.rdf.model.Resource asResource(Model model) {
        return model.createResource(getHyperlink());
    }

    @Deprecated
    public String getAnalyzeHyperlink() {
        return deeplinker.getHost() + "/rest/analyze?deeplink=" + DeepLinker.encode(getHyperlink());
    }

    @Override
    public int hashCode() {
        int hash = 3;
        hash = 31 * hash + Objects.hashCode(getHyperlink());
        return hash;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final DeepLink other = (DeepLink) obj;
        if (!Objects.equals(getHyperlink(), other.getHyperlink())) {
            return false;
        }
        return true;
    }

    public DeepLink clearResources() {
        this.resources.clear();
        return this;
    }

    public URI toURI() {
        return URI.create(getHyperlink());
    }
}
