package de.dfki.sds.hephaistos.deeplink;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.lang.reflect.Method;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.Map.Entry;
import java.util.StringJoiner;
import org.apache.jena.rdf.model.Model;
import org.apache.jena.shared.PrefixMapping;
import org.apache.jena.shared.impl.PrefixMappingImpl;

/**
 *
 * @author Markus Schr&ouml;der
 */
public class DeepLinker {

    //host where DeepLinker runs
    private String host;

    //root entry points (configured in main)
    private RMap root;

    private static final String fiLocation = "de.dfki.sds.hephaistos.deeplink.";

    public DeepLinker(String host) {
        this.host = host;

        //configurable entry point
        root = new RMap();
    }

    public DeepLink processURI(String uri) throws Exception {
        return processURI(uri, false);
    }

    public DeepLink processURI(String uri, boolean skipResources) throws Exception {
        if (!uri.startsWith(host)) {
            throw new RuntimeException("host not match");
        }

        return processPath(decode(uri.substring(host.length())), skipResources);
    }

    public DeepLink processPath(String path) throws Exception {
        return processPath(path, false);
    }

    //skipResources: skips the actual processing
    public DeepLink processPath(String path, boolean skipResources) throws Exception {
        DeepLink ctx = new DeepLink(this);
        ctx.getResources().add(root);

        String[] parts = path.split("/");

        //System.out.println(path);
        for (String part : parts) {
            if (part.trim().isEmpty()) {
                continue;
            }

            //part = decode(part);
            //String origPart = part;
            int colonIndex = part.indexOf(DeepLink.SEPARATOR);

            //jörn suggestion: if nothing is given assume child frag id
            if (colonIndex == -1) {
                part = "child@" + part;
                colonIndex = part.indexOf(DeepLink.SEPARATOR);
            }

            String[] nameParams = new String[]{part.substring(0, colonIndex), part.substring(colonIndex + 1, part.length())};
            String[] strparams = nameParams[1].split(",");
            for (int i = 0; i < strparams.length; i++) {
                strparams[i] = decode(strparams[i]);
            }
            Object[] params = new Object[strparams.length];
            System.arraycopy(strparams, 0, params, 0, params.length);
            boolean dontCare = params.length == 1 && params[0].equals(DeepLink.DONTCARE);
            if (dontCare) {
                params = new String[0];
            }
            String name = nameParams[0];
            boolean isTo = name.equals("to");

            String className;
            if (!isTo) {
                String nameProper = name.substring(0, 1).toUpperCase() + name.substring(1);
                className = fiLocation + "FI" + nameProper;
            } else {
                String nameProper = name.substring(0, 1).toUpperCase() + name.substring(1);
                String param1Proper = ((String) params[0]).substring(0, 1).toUpperCase() + ((String) params[0]).substring(1);
                name = (String) params[0];
                //rest params
                params[0] = ctx;
                //params = Arrays.copyOfRange(params, 1, params.length);
                //params = new String[0];
                className = fiLocation + "FI" + nameProper + param1Proper;
            }

            if (!skipResources) {

                Class<?> fiClass = this.getClass().getClassLoader().loadClass(className);

                DLResource last = ctx.last();
                
                

                if (!fiClass.isAssignableFrom(last.getClass())) {
                    throw new RuntimeException(last + " does not implement " + className);
                }

                Class[] paramsTypes = new Class[params.length];
                for (int i = 0; i < params.length; i++) {
                    paramsTypes[i] = params[i].getClass();
                }

                //special case give the current path processing result
                /*
                if(isTo) {
                    paramsTypes = new Class[1];
                    paramsTypes[0] = PathProcessingResult.class;
                    params = new Object[] { ctx };
                }
                 */
                Method method = fiClass.getMethod(name, paramsTypes);

                DLResource resultResource = (DLResource) method.invoke(last, params);
                
                if(resultResource == null) {
                    throw new RuntimeException(part + " not found in " + path + " (null was returned)");
                }
                
                ctx.getResources().add(resultResource);
            }

            ctx.getMethods().add(name);
            ctx.getParams().add(params);

            //build fragid together with correct encoding
            part = "";

            if (isTo) {
                part += "to";
            } else if (!name.equals("child")) {
                part += name;
            }

            if (isTo || !name.equals("child")) {
                part += DeepLink.SEPARATOR;
            }

            StringJoiner sj = new StringJoiner(DeepLink.SEPARATOR_PARAM);
            for (Object param : params) {
                if (param instanceof DeepLink) {
                    sj.add(name);
                }
                if (param instanceof String) {
                    //always encoded
                    sj.add(encode2((String) param));
                }
            }
            if (params.length == 0) {
                part += "*";
            }
            //params
            part += sj.toString();
            //}
            ctx.getFragids().add(part);

            //save it
            //part = origPart;
            //always encode?
            //if(part.contains("/") || part.contains(">") || part.contains("#") || part.contains("(") || part.contains(")")) {
            //has to be twice
            //part = URLEncoder.encode(part, "UTF-8");
            //part = URLEncoder.encode(part, "UTF-8");
            //}
            //System.out.println("\t" + last + " " + name + Arrays.toString(params) + " " + resultResource);
        }

        return ctx;
    }

    //<editor-fold desc="vocab">
    public PrefixMapping getPrefixMapping() {
        PrefixMapping prefixmapping = new PrefixMappingImpl();
        prefixmapping.setNsPrefixes(PrefixMapping.Extended);
        prefixmapping.setNsPrefix("foaf", "http://xmlns.com/foaf/0.1/");
        //if (hasVocab()) {
        //    prefixmapping.setNsPrefix("dl", getVocabEndpoint() + "/");
        //}
        return prefixmapping;
    }

    public String getPrefixMappingForSPARQL() {
        String sparql = "";
        for (Entry<String, String> e : getPrefixMapping().getNsPrefixMap().entrySet()) {
            sparql += String.format("prefix %s: <%s>\n", e.getKey(), e.getValue());
        }
        return sparql;
    }

    //</editor-fold>
    public String getHost() {
        return host;
    }

    public RMap getRoot() {
        return root;
    }

    //helper ===================================================================
    
    public static String decode(String part) {
        try {
            return URLDecoder.decode(part, "UTF-8");
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public static String encode(String part) {
        try {
            return URLEncoder.encode(part, "UTF-8");
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public static String encode2(String part) {
        return encode(encode(part));
    }

    public static String toTurtle(Model model) {
        StringWriter sw = new StringWriter();
        model.write(sw, "TTL");
        return sw.toString();
    }

    public static void printTurtle(Model model) {
        System.out.println(toTurtle(model));
    }

    public static void writeTurtle(Model model, File f) {
        try (FileWriter fw = new FileWriter(f)) {
            fw.write(toTurtle(model));
        } catch (IOException ex) {
            throw new RuntimeException(ex);
        }
    }

}
