package de.dfki.sds.hephaistos.storage;

import de.dfki.sds.mschroeder.commons.lang.MapUtility;
import de.dfki.sds.mschroeder.commons.lang.data.Couple;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.rdf.model.Property;
import org.apache.jena.rdf.model.RDFNode;
import org.apache.jena.rdf.model.Resource;
import org.apache.jena.rdf.model.ResourceFactory;
import org.apache.jena.rdf.model.Statement;

/**
 * Maps from A to B and vice versa (in RAM).
 * Implementations override methods for customization.
 * It can read and write to RDF to persist the mapping.
 * @author Markus Schr&ouml;der
 */
public abstract class Mapper<A, B> {
    
    /**
     * The main property that is used for RDF persistence.
     */
    public static final Property isMappedTo = ResourceFactory.createProperty(Utils.URI_SCHEME + ":isMappedTo");
    
    //private Set<Couple<A, B>> links;
    
    private Map<A, List<B>> a2bs;
    private Map<B, List<A>> b2as;
    
    public Mapper() {
        //links = new HashSet<>();
        a2bs = new HashMap<>();
        b2as = new HashMap<>();
    }
    
    public void add(A a, B b) {
        nullCheck(a, b);
        //links.add(new Couple<>(a,b));
        MapUtility.fillKey2ListMap(a, b, a2bs);
        MapUtility.fillKey2ListMap(b, a, b2as);
        
        //TODO sort it
        //a2bs.get(a).sort(com);
    }
    
    public void remove(A a, B b) {
        nullCheck(a, b);
        //links.remove(new Couple<>(a,b));
        
        List<B> bs = a2bs.get(a);
        bs.remove(b);
        if(bs.isEmpty()) {
            a2bs.remove(a);
        }
        
        List<A> as = b2as.get(b);
        as.remove(a);
        if(as.isEmpty()) {
            b2as.remove(b);
        }
    }
    
    public void removeA(A a) {
        a2bs.remove(a);
        for(Object eobj : b2as.entrySet().toArray()) {
            Entry<B, List<A>> e = (Entry<B, List<A>>) eobj;
            e.getValue().remove(a);
            if(e.getValue().isEmpty()) {
                b2as.remove(e.getKey());
            }
        }
    }
    
    public void removeB(B b) {
        b2as.remove(b);
        for(Object eobj : a2bs.entrySet().toArray()) {
            Entry<A, List<B>> e = (Entry<A, List<B>>) eobj;
            e.getValue().remove(b);
            if(e.getValue().isEmpty()) {
                a2bs.remove(e.getKey());
            }
        }
    }
    
    private void nullCheck(A a, B b) {
        if(a == null)
            throw new RuntimeException("a is null, maybe check fromUriA method impl");
        
        if(b == null)
            throw new RuntimeException("b is null, maybe check fromUriB method impl");
    }
    
    public List<B> getBsOf(A a) {
        return a2bs.getOrDefault(a, Arrays.asList());
        //return links.stream().filter(c -> c.getLeft().equals(a)).map(c -> c.getRight()).collect(toSet());
    }
    
    /*
    public List<B> getBsOf(A a, Comparator<B> comparator) {
        List<B> l = new ArrayList<>(getBsOf(a));
        l.sort(comparator);
        return l;
    }
    */
    
    public List<A> getAsOf(B b) {
        return b2as.getOrDefault(b, Arrays.asList());
        //return links.stream().filter(c -> c.getRight().equals(b)).map(c -> c.getLeft()).collect(toSet());
    }
    
    /*
    public List<A> getAsOf(B b, Comparator<A> comparator) {
        List<A> l = new ArrayList<>(getAsOf(b));
        l.sort(comparator);
        return l;
    }
    */
    
    public abstract String toUriA(A a);
    
    public abstract String toUriB(B b);
    
    //public abstract A fromUriA(String uri);
    
    //public abstract B fromUriB(String uri);
    
    public abstract Map<String, A> fromUrisA(List<String> uris);
    
    public abstract Map<String, B> fromUrisB(List<String> uris);
    
    public Model getModel() {
        Model model = ModelFactory.createDefaultModel();
        for(Couple<A,B> link : getLinkList()) {
            Resource a = ResourceFactory.createResource(toUriA(link.getLeft()));
            Resource b = ResourceFactory.createResource(toUriB(link.getRight()));
            model.add(a, isMappedTo, b);
        }
        return model;
    }
    
    public void clear() {
        a2bs.clear();
        b2as.clear();
    }
    
    public void fromModel(Model model) {
        fromModel(model, true);
    }
    
    public void fromModel(Model model, boolean skipLinksHavingNull) {
        
        //all distinct As
        List<String> urisA = new ArrayList<>();
        for(Resource res : model.listSubjectsWithProperty(isMappedTo).toSet()) {
            urisA.add(res.getURI());
        }
        
        //all distinct Bs
        List<String> urisB = new ArrayList<>();
        for(RDFNode res : model.listObjectsOfProperty(isMappedTo).toSet()) {
            urisB.add(res.asResource().getURI());
        }
        
        //A
        Map<String, A> uri2a = fromUrisA(urisA);
        /*
        Map<String, A> uri2a = new HashMap<>();
        for(int i = 0; i < urisA.size(); i++) {
            uri2a.put(urisA.get(i), as.get(i));
        }
        */
        //B
        Map<String, B> uri2b = fromUrisB(urisB);
        
        /*
        Map<String, B> uri2b = new HashMap<>();
        for(int i = 0; i < urisB.size(); i++) {
            uri2b.put(urisB.get(i), bs.get(i));
        }
        */
        
        for(Statement stmt : model.listStatements(null, isMappedTo, (RDFNode) null).toList()) {
            
            A a = uri2a.get(stmt.getSubject().getURI());
            B b = uri2b.get(stmt.getObject().asResource().getURI());
            
            if(!skipLinksHavingNull || (a != null && b != null)) {
                add(a, b);
            }
        }
    }
    
    public List<Couple<A, B>> getLinkList() {
        List<Couple<A, B>> l = new ArrayList<>();
        
        for(A a : a2bs.keySet()) {
            for(B b : a2bs.get(a)) {
                l.add(new Couple<>(a,b));
            }
        }
        
        /*
        l.sort((o1, o2) -> {
            return Integer.compare(o1.getLeft().hashCode(), o2.getLeft().hashCode());
        });
        */
        
        return l;
    }
    
    public int size() {
        return getLinkList().size();
    }
}
