package de.dfki.sds.hephaistos.storage.dsv;

import de.dfki.sds.hephaistos.DataStoreDescription;
import de.dfki.sds.hephaistos.Preference;
import de.dfki.sds.hephaistos.Setting;
import de.dfki.sds.hephaistos.storage.DataIO;
import de.dfki.sds.hephaistos.storage.DataModel;
import de.dfki.sds.hephaistos.storage.InternalStorage;
import de.dfki.sds.hephaistos.storage.StorageManager;
import de.dfki.sds.mschroeder.commons.lang.FileUtility;
import de.dfki.sds.mschroeder.commons.lang.swing.LoadingListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.zip.GZIPInputStream;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableModel;
import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVParser;
import org.apache.commons.csv.CSVRecord;

/**
 *
 * @author Markus Schr&ouml;der
 */
public class DSVStorageIO extends DataIO<DSVStorage, TableModel> {

    @Override
    public InternalStorage createInternalStorage(StorageManager storageManager) {
        return storageManager.getDSVStorage();
    }
    
    @Override
    public void exporting(DSVStorage from, DataStoreDescription to, LoadingListener listener) {
        throw new RuntimeException("not implemented yet");
    }

    @Override
    public void importing(DataStoreDescription from, DSVStorage to, LoadingListener listener) {
        CSVFormat format = getFormat(from);
        try {
            for(String locator : from.getLocators()) {
                CSVParser parser = getParser(locator, from.getPreference(), format);
                Iterator<CSVRecord> iter = parser.iterator();
                
                int offset = from.getPreference().getValueAsInt("offset", 0);
                int limit = from.getPreference().getValueAsInt("limit", -1);
                int bulkLimit = from.getPreference().getValueAsInt("bulk size", 50000);
                
                if(listener != null) {
                    if(limit < 0) {
                        //locator points to a file
                        listener.setMaximum(FileUtility.countLines(new File(locator)));
                    } else {
                        listener.setMaximum(limit);
                    }
                }
                
                for(int current = 0; limit == -1 ? iter.hasNext() : current < limit; current += bulkLimit) {
                    
                    List<DSVRecord> bulk = getBulk(iter, from.getParts(), current, offset, current + bulkLimit);
                    to.addBulk(bulk);
                    
                    if(listener != null) {
                        listener.setCurrent(current);
                    }
                    
                    if(listener.cancel()) {
                        break;
                    }
                    
                    //10000 bluk limit: 45,2 MB used / 251,7 MB total / 3,7 GB max
                    //100000 bluk limit: 954,7 MB used / 1,1 GB total / 3,7 GB max
                    //lamma table has row count 4247217
                    //System.out.println(bulkLimit + " bulk limit, " + MemoryUtility.memoryStatistics());
                }
                
                parser.close();
            }
        } catch (IOException ex) {
            throw new RuntimeException(ex);
        }
    }
    
    private CSVFormat getFormat(DataStoreDescription from) {
        CSVFormat format = CSVFormat.DEFAULT;
        
        Preference pref = from.getPreference();
        
        boolean tab = pref.getValueAsBoolean("delimiter is tab", false);
        format = format.withDelimiter(tab ? '\t' : pref.getValueAsString("delimiter", ",").charAt(0));
        
        boolean firstRecordAsHeader = pref.getValueAsBoolean("first record as header", false);
        if(firstRecordAsHeader) {
            format = format.withFirstRecordAsHeader();
        }
        
        return format;
    }

    private List<DSVRecord> loadBulk(String locator, List<String> parts, Preference pref, CSVFormat format) throws IOException {
        CSVParser parser = getParser(locator, pref, format);

        Iterator<CSVRecord> iter = parser.iterator();

        int current = 0;
        int offset = pref.getValueAsInt("offset", 0);
        int limit = pref.getValueAsInt("limit", -1);
        
        List<DSVRecord> bulk = getBulk(iter, parts, current, offset, limit);
        
        parser.close();
        
        return bulk;
    }
    
    private CSVParser getParser(String locator, Preference pref, CSVFormat format) throws IOException {
        InputStream is = new FileInputStream(locator);
                
        if(pref.getValueAsBoolean(Setting.GZIP, false)) {
            is = new GZIPInputStream(is);
        }
        
        Reader reader = new InputStreamReader(is, pref.getValueAsString(Setting.CHARSET, "UTF-8"));

        CSVParser parser = format.parse(reader);
        
        return parser;
    }
    
    private List<DSVRecord> getBulk(Iterator<CSVRecord> iter, List<String> parts, int current, int offset, int limit) {
        DSVHeader header = null;
        List<DSVRecord> bulk = new ArrayList<>();
        while(iter.hasNext()) {
            CSVRecord csvRecord = iter.next();

            if(header == null) {
                if(DSVHeader.hasHeader(csvRecord)) {
                    header = new DSVHeader(csvRecord);
                } else if(!parts.isEmpty()) {
                    header = new DSVHeader(parts);
                }
            }

            if(current >= offset) {
                bulk.add(new DSVRecord(header, csvRecord));
            }
            
            current++;
            
            if(limit >= 0 && current >= limit) {
                break;
            }
        }
        return bulk;
    }
    
    @Override
    public TableModel preview(DataStoreDescription from) {
        
        Object[][] data = new Object[0][0];
        Object[] columnNames = new Object[0];
        
        CSVFormat format = getFormat(from);
        
        Preference prefCopy = new Preference(from.getPreference());
        prefCopy.setValue("limit", 5);
        
        try {
            List<DSVRecord> bulk = loadBulk(from.getLocators().get(0), from.getParts(), prefCopy, format);
            
            if(bulk.size() > 0) {
            
                //header
                if(bulk.get(0).hasHeader()) {
                    columnNames = bulk.get(0).getHeader().toList().toArray();
                } else {
                    columnNames = DSVStorageSqlite.getVirtualColumnNames(bulk.get(0)).toArray();
                }
                
                //data
                data = new Object[bulk.size()][];
                int i = 0;
                for(DSVRecord record : bulk) {
                    data[i] = record.getValues();
                    i++;
                }
            }
            
        } catch (IOException ex) {
            throw new RuntimeException(ex);
        }
        
        DefaultTableModel dtm = new DefaultTableModel(data, columnNames);
        return dtm;
    }

    @Override
    public Preference getPreference() {
        return new Preference(
                new Setting("delimiter", ","),
                new Setting("delimiter is tab", false),
                new Setting("first record as header", true),
                new Setting("offset", 0),
                new Setting("limit", -1),
                new Setting("bulk size", 50000),
                new Setting(Setting.GZIP, false),
                new Setting(Setting.CHARSET, "UTF-8")
        );
    }

    @Override
    public String getName() {
        return "Delimiter-Separated Value";
    }

    @Override
    public DataModel getDataModel() {
        return DataModel.Relational;
    }
    
}
