package de.dfki.sds.hephaistos.storage.dsv;

import de.dfki.sds.hephaistos.storage.InternalStorageMetaData;
import de.dfki.sds.hephaistos.storage.ResultSetIterator;
import de.dfki.sds.hephaistos.storage.SQLiteUtility;
import de.dfki.sds.hephaistos.storage.StorageManager;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.StringJoiner;
import org.apache.commons.io.IOUtils;

/**
 *
 * @author Markus Schr&ouml;der
 */
public class DSVStorageSqlite extends DSVStorage {

    private Connection connection;
    private String tablename;

    public DSVStorageSqlite(InternalStorageMetaData metaData, Connection connection) {
        super(metaData);
        this.tablename = metaData.getId();
        this.connection = connection;
    }

    private boolean tableExists() {
        return SQLiteUtility.supply(connection, c -> {
            DatabaseMetaData dbm = c.getMetaData();
            // check if "employee" table is there
            ResultSet tables = dbm.getTables(null, null, tablename, null);
            return tables.next();
        });
    }

    public static List<String> getVirtualColumnNames(DSVRecord record) {
        if (record.size() > 26) {
            throw new RuntimeException("column name generation problem");
        }

        List<String> l = new ArrayList<>();

        //just use column names: A B C D E F ...
        for (int i = 0; i < record.size(); i++) {
            l.add(String.valueOf((char) ('A' + i)));
        }

        return l;
    }

    private String getQuery(String path) {
        return getQuery(path, null);
    }

    private String getQuery(String path, String columns) {
        try {
            String query = IOUtils.toString(getClass().getResourceAsStream(path), StandardCharsets.UTF_8);
            query = query.replace("${tablename}", tablename);
            if (columns != null) {
                query = query.replace("${columns}", columns);
            }
            return query;
        } catch (IOException ex) {
            throw new RuntimeException(ex);
        }
    }

    private void createTable(DSVRecord record) {
        StringJoiner sj = new StringJoiner(",");
        //"name" TEXT
        if (record.hasHeader()) {
            record.getHeader().toList().forEach(col -> sj.add("\"" + col + "\" TEXT"));
        } else {
            getVirtualColumnNames(record).forEach(col -> sj.add("\"" + col + "\" TEXT"));
        }
        String columns = sj.toString();

        SQLiteUtility.execute(connection, getQuery("/de/dfki/sds/hephaistos/storage/dsv/CreateTable.sql", columns));
    }

    @Override
    public void clear() {
        if (tableExists()) {
            SQLiteUtility.execute(connection, getQuery("/de/dfki/sds/hephaistos/storage/dsv/DeleteTable.sql"));
        }
    }

    @Override
    public void remove() {
        SQLiteUtility.execute(connection, getQuery("../Drop.sql"));
    }

    @Override
    public DSVStorageSummary summary() {
        return new DSVStorageSummary(this, metaData.getSummaryCache());
    }

    /*
    @Override
    public List<DSVRecord> search(String keywords) {
        throw new RuntimeException("not implemented yet");
    }

    @Override
    public ResultSet query(String query) {
        throw new RuntimeException("not implemented yet");
    }

    @Override
    public void execute(String query) {
        throw new RuntimeException("not implemented yet");
    }
    */
    
    @Override
    public void close() {
        //ignore
    }

    @Override
    public void addBulk(Collection<DSVRecord> items) {
        if (items.isEmpty()) {
            return;
        }

        DSVRecord first = items.iterator().next();

        if (!tableExists()) {
            createTable(first);
        }

        int headerSize;
        if (first.hasHeader()) {
            headerSize = first.getHeader().size();
        } else {
            //all
            headerSize = first.size();
        }

        //? columns
        StringJoiner sj = new StringJoiner(",");
        for (int i = 0; i < headerSize; i++) {
            sj.add("?");
        }
        String columns = sj.toString();

        String query = getQuery("/de/dfki/sds/hephaistos/storage/dsv/InsertRecord.sql", columns);

        SQLiteUtility.run(connection, c -> {

            c.setAutoCommit(false);

            PreparedStatement ps = c.prepareStatement(query);

            for (DSVRecord record : items) {
                if (record.getRecordNumber() < 0) {
                    ps.setNull(1, Types.BIGINT);
                } else {
                    ps.setLong(1, record.getRecordNumber());
                }
                ps.setLong(2, record.getCharacterPosition());
                ps.setString(3, record.getComment());

                for (int i = 0; i < headerSize; i++) {
                    if(i < record.getValues().length) {
                        ps.setString(4 + i, record.getValues()[i]);
                    } else {
                        //header is larger than record length
                        ps.setNull(4 + i, Types.VARCHAR);
                    }
                }

                ps.addBatch();
            }

            ps.executeBatch();
            
            c.commit();
            
            ps.close();

            c.setAutoCommit(true);
        });

    }

    @Override
    public void removeBulk(Collection<DSVRecord> items) {
        if (items.isEmpty()) {
            return;
        }

        StringJoiner sj = new StringJoiner(",");
        for (DSVRecord record : items) {
            if (record.getRecordNumber() < 0) {
                continue;
            }
            sj.add(String.valueOf(record.getRecordNumber()));
        }
        String ids = sj.toString();

        String query = getQuery("/de/dfki/sds/hephaistos/storage/dsv/DeleteRecord.sql");
        query = query.replace("${ids}", ids);
        SQLiteUtility.execute(connection, query);
    }

    //because it is an iterable we can not synchronize the connection
    @Override
    public Iterable<DSVRecord> getListIter() {

        return () -> {
            if (!tableExists()) {
                return new Iterator<DSVRecord>() {
                    @Override
                    public boolean hasNext() {
                        return false;
                    }

                    @Override
                    public DSVRecord next() {
                        throw new NoSuchElementException();
                    }
                };
            }
            
            //SQLiteUtility not usable because iterator will return
            String query = getQuery("/de/dfki/sds/hephaistos/storage/dsv/SelectRecord.sql");
            PreparedStatement ps;
            try {
                ps = connection.prepareStatement(query);
            } catch (SQLException ex) {
                throw new RuntimeException(ex);
            }

            //PreparedStatement will be closed by the iterator
            return new ResultSetIterator<>(ps, rs -> {

                DSVHeader header = new DSVHeader();
                DSVRecord record = new DSVRecord();

                try {
                    record.setDeepLink(StorageManager.DEEPLINK + "/" + getId() + "/" + rs.getInt("id"));
                    
                    //header
                    ResultSetMetaData m = rs.getMetaData();
                    for (int i = 4; i <= m.getColumnCount(); i++) {
                        header.put(m.getColumnName(i), i - 4);
                    }
                    record.setHeader(header);
                    
                    //meta data
                    record.setRecordNumber(rs.getLong(1));
                    record.setCharacterPosition(rs.getLong(2));
                    record.setComment(rs.getString(3));

                    //values
                    int size = rs.getMetaData().getColumnCount() - 3;
                    String[] values = new String[size];
                    for (int i = 0; i < size; i++) {
                        values[i] = rs.getString(4 + i);
                    }
                    record.setValues(values);
                    
                } catch (SQLException ex) {
                    throw new RuntimeException(ex);
                }

                return record;
            });
        };
    }

    @Override
    public long size() {
        if(!tableExists())
            return 0;
        
        String query = getQuery("../Count.sql");
        return SQLiteUtility.supply(connection, c -> {
            PreparedStatement stmt = c.prepareStatement(query);
            ResultSet rs = stmt.executeQuery();
            long size = rs.getLong(1);
            stmt.close();
            return size;
        });
    }

    /*package*/ String getTablename() {
        return tablename;
    }

    @Override
    public List<String> listDeepLinks() {
        String query = getQuery("/de/dfki/sds/hephaistos/storage/dsv/SelectId.sql");
        
        List<String> result = new ArrayList<>();
        
        SQLiteUtility.run(connection, (c) -> {
            PreparedStatement ps = c.prepareStatement(query);
            ResultSet rs = ps.executeQuery();
            while(rs.next()) {
                result.add(StorageManager.DEEPLINK + "/" + getId() + "/" + rs.getInt(1));
            }
            ps.close();
        });
        
        return result;
    }

}
