package de.dfki.sds.hephaistos.storage.excel;

import de.dfki.sds.hephaistos.storage.InternalStorageMetaData;
import de.dfki.sds.hephaistos.storage.ResultSetIterator;
import de.dfki.sds.hephaistos.storage.SQLiteUtility;
import de.dfki.sds.hephaistos.storage.StorageManager;
import java.awt.Color;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.StringJoiner;
import org.apache.commons.io.IOUtils;

/**
 * Implementation in sqlite.
 *
 * @author Markus Schr&ouml;der
 */
public class ExcelStorageSqlite extends ExcelStorage {

    private Connection connection;
    private String tablename;

    public ExcelStorageSqlite(InternalStorageMetaData metaData, Connection connection) {
        super(metaData);
        this.connection = connection;
        this.tablename = metaData.getId();
        createTables();
    }

    private String getQuery(String path) {
        try {
            String query = IOUtils.toString(getClass().getResourceAsStream(path), StandardCharsets.UTF_8);
            query = query.replaceAll("\\$\\{tablename\\}", tablename);
            return query;
        } catch (IOException ex) {
            throw new RuntimeException(ex);
        }
    }

    private void createTables() {
        SQLiteUtility.execute(connection, getQuery("/de/dfki/sds/hephaistos/storage/excel/CreateTableSheet.sql"));
        SQLiteUtility.execute(connection, getQuery("/de/dfki/sds/hephaistos/storage/excel/CreateTableCell.sql"));
        SQLiteUtility.execute(connection, getQuery("/de/dfki/sds/hephaistos/storage/excel/CreateTableTextStyle.sql"));
    }

    @Override
    public void addBulk(Collection<ExcelSheet> items) {
        String query = getQuery("/de/dfki/sds/hephaistos/storage/excel/InsertSheet.sql");

        SQLiteUtility.run(connection, c -> {
            c.setAutoCommit(false);

            PreparedStatement ps = c.prepareStatement(query);

            for (ExcelSheet sheet : items) {

                ps.setInt(1, sheet.getId());
                ps.setString(2, sheet.getName());
                ps.setInt(3, sheet.getWidth());
                ps.setInt(4, sheet.getHeight());
                ps.setString(5, sheet.getFilename());
                ps.setString(6, sheet.getFilePath());

                ps.addBatch();
            }

            ps.executeBatch();
            c.commit();
            ps.close();
            c.setAutoCommit(true);
        });
    }

    @Override
    public void addCellBulk(Collection<ExcelCell> items) {
        String query = getQuery("/de/dfki/sds/hephaistos/storage/excel/InsertCell.sql");

        SQLiteUtility.run(connection, c -> {
            c.setAutoCommit(false);

            PreparedStatement ps = c.prepareStatement(query);

            for (ExcelCell cell : items) {

                ps.setInt(1, cell.getId());
                ps.setInt(2, cell.getSheetId());
                ps.setInt(3, cell.getRow());
                ps.setInt(4, cell.getColumn());
                ps.setString(5, cell.getAddress());

                ps.setBoolean(6, cell.isFontBold());
                ps.setBoolean(7, cell.isFontItalic());
                ps.setBoolean(8, cell.isFontStrikeout());
                ps.setBoolean(9, cell.isFontUnderline());
                ps.setString(10, cell.getFontName());
                ps.setString(11, color2string(cell.getFontColor()));

                ps.setString(12, color2string(cell.getForegroundColor()));
                ps.setString(13, color2string(cell.getBackgroundColor()));
                ps.setInt(14, cell.getRotation());
                ps.setString(15, cell.getHorizontalAlignment());
                ps.setString(16, cell.getVerticalAlignment());

                ps.setString(17, cell.getBorderTop());
                ps.setString(18, cell.getBorderLeft());
                ps.setString(19, cell.getBorderRight());
                ps.setString(20, cell.getBorderBottom());

                ps.setString(21, cell.getCellType());
                ps.setString(22, cell.getCellTypeFormular());

                ps.setDouble(23, cell.getValueNumeric());
                ps.setBoolean(24, cell.getValueBoolean());
                ps.setString(25, cell.getValueFormular());
                ps.setInt(26, cell.getValueError());
                ps.setString(27, cell.getValueString());

                ps.setInt(28, cell.getFontSize());
                ps.setString(29, cell.getValueRichText());
                
                ps.addBatch();
            }

            ps.executeBatch();
            c.commit();
            ps.close();
            c.setAutoCommit(true);
        });
    }

    @Override
    public void addTextStyleBulk(Collection<ExcelTextStyle> items) {
        String query = getQuery("/de/dfki/sds/hephaistos/storage/excel/InsertTextStyle.sql");

        SQLiteUtility.run(connection, c -> {
            c.setAutoCommit(false);

            PreparedStatement ps = c.prepareStatement(query);

            for (ExcelTextStyle textStyle : items) {

                ps.setInt(1, textStyle.getId());
                ps.setInt(2, textStyle.getCellId());
                
                ps.setInt(3, textStyle.getBegin());
                ps.setInt(4, textStyle.getEnd());
                ps.setString(5, textStyle.getText());

                ps.setBoolean(6, textStyle.isFontBold());
                ps.setBoolean(7, textStyle.isFontItalic());
                ps.setBoolean(8, textStyle.isFontStrikeout());
                ps.setBoolean(9, textStyle.isFontUnderline());
                ps.setString(10, textStyle.getFontName());
                ps.setString(11, color2string(textStyle.getFontColor()));
                ps.setInt(12, textStyle.getFontSize());

                ps.addBatch();
            }

            ps.executeBatch();
            c.commit();
            ps.close();
            c.setAutoCommit(true);
        });
    }
    
    @Override
    public void removeBulk(Collection<ExcelSheet> items) {
        if (items.isEmpty()) {
            return;
        }

        StringJoiner sj = new StringJoiner(",");
        for (ExcelSheet sheet : items) {
            sj.add(String.valueOf(sheet.getId()));
        }
        String ids = sj.toString();

        String query = getQuery("/de/dfki/sds/hephaistos/storage/excel/DeleteSheet.sql");
        query = query.replace("${ids}", ids);
        SQLiteUtility.execute(connection, query);
    }

    @Override
    public void removeCellBulk(Collection<ExcelCell> items) {
        if (items.isEmpty()) {
            return;
        }

        StringJoiner sj = new StringJoiner(",");
        for (ExcelCell cell : items) {
            sj.add(String.valueOf(cell.getId()));
        }
        String ids = sj.toString();

        String query = getQuery("/de/dfki/sds/hephaistos/storage/excel/DeleteCell.sql");
        query = query.replace("${ids}", ids);
        SQLiteUtility.execute(connection, query);
    }
   
    @Override
    public void removeTextStyleBulk(Collection<ExcelTextStyle> items) {
        if (items.isEmpty()) {
            return;
        }

        StringJoiner sj = new StringJoiner(",");
        for (ExcelTextStyle textStyle : items) {
            sj.add(String.valueOf(textStyle.getId()));
        }
        String ids = sj.toString();

        String query = getQuery("/de/dfki/sds/hephaistos/storage/excel/DeleteTextStyle.sql");
        query = query.replace("${ids}", ids);
        SQLiteUtility.execute(connection, query);
    }

    
    
    @Override
    public Iterable<ExcelSheet> getListIter() {
        return () -> {
            String query = getQuery("/de/dfki/sds/hephaistos/storage/excel/SelectSheet.sql");
            PreparedStatement ps;
            try {
                ps = connection.prepareStatement(query);
            } catch (SQLException ex) {
                throw new RuntimeException(ex);
            }

            return new ResultSetIterator<>(ps, rs -> {
                return resultSetToSheet(rs);
            });
        };
    }

    private ExcelSheet resultSetToSheet(ResultSet rs) {
        ExcelSheet sheet = new ExcelSheet();
        try {
            sheet.setDeepLink(StorageManager.DEEPLINK + "/" + getId() + "/sheet/" + rs.getInt(1));
            
            sheet.setId(rs.getInt(1));

            sheet.setName(rs.getString(2));

            sheet.setWidth(rs.getInt(3));
            sheet.setHeight(rs.getInt(4));

            sheet.setFilename(rs.getString(5));
            sheet.setFilePath(rs.getString(6));
        } catch (SQLException ex) {
            throw new RuntimeException(ex);
        }
        return sheet;
    }

    private ExcelCell resultSetToCell(ResultSet rs) {
        ExcelCell cell = new ExcelCell();
        try {
            cell.setDeepLink(StorageManager.DEEPLINK + "/" + getId() + "/cell/" + rs.getInt(1));
            
            cell.setId(rs.getInt(1));
            cell.setSheetId(rs.getInt(2));

            cell.setRow(rs.getInt(3));
            cell.setColumn(rs.getInt(4));
            cell.setAddress(rs.getString(5));

            cell.setFontBold(rs.getBoolean(6));
            cell.setFontItalic(rs.getBoolean(7));
            cell.setFontStrikeout(rs.getBoolean(8));
            cell.setFontUnderline(rs.getBoolean(9));

            cell.setFontName(rs.getString(10));
            cell.setFontColor(string2color(rs.getString(11)));

            cell.setForegroundColor(string2color(rs.getString(12)));
            cell.setBackgroundColor(string2color(rs.getString(13)));
            cell.setRotation(rs.getInt(14));
            cell.setHorizontalAlignment(rs.getString(15));
            cell.setVerticalAlignment(rs.getString(16));

            cell.setBorderTop(rs.getString(17));
            cell.setBorderLeft(rs.getString(18));
            cell.setBorderRight(rs.getString(19));
            cell.setBorderBottom(rs.getString(20));

            cell.setCellType(rs.getString(21));
            cell.setCellTypeFormular(rs.getString(22));

            cell.setValueNumeric(rs.getDouble(23));
            cell.setValueBoolean(rs.getBoolean(24));
            cell.setValueFormular(rs.getString(25));
            cell.setValueError(rs.getInt(26));
            cell.setValueString(rs.getString(27));
            
            cell.setFontSize(rs.getInt(28));
            cell.setValueRichText(rs.getString(29));

        } catch (SQLException ex) {
            throw new RuntimeException(ex);
        }
        return cell;
    }
    
    private ExcelTextStyle resultSetToTextStyle(ResultSet rs) {
        ExcelTextStyle textStyle = new ExcelTextStyle();
        try {
            textStyle.setDeepLink(StorageManager.DEEPLINK + "/" + getId() + "/textStyle/" + rs.getInt(1));
            
            textStyle.setId(rs.getInt(1));
            textStyle.setCellId(rs.getInt(2));
            
            textStyle.setBegin(rs.getInt(3));
            textStyle.setEnd(rs.getInt(4));
            textStyle.setText(rs.getString(5));
            
            textStyle.setFontBold(rs.getBoolean(6));
            textStyle.setFontItalic(rs.getBoolean(7));
            textStyle.setFontStrikeout(rs.getBoolean(8));
            textStyle.setFontUnderline(rs.getBoolean(9));

            textStyle.setFontName(rs.getString(10));
            textStyle.setFontColor(string2color(rs.getString(11)));
            textStyle.setFontSize(rs.getInt(12));

        } catch (SQLException ex) {
            throw new RuntimeException(ex);
        }
        return textStyle;
    }

    @Override
    public Iterable<ExcelCell> getCellListIter(ExcelSheet sheet) {
        return () -> {
            String query = getQuery("/de/dfki/sds/hephaistos/storage/excel/SelectCell.sql");
            PreparedStatement ps;
            try {
                ps = connection.prepareStatement(query);
                ps.setInt(1, sheet.getId());
            } catch (SQLException ex) {
                throw new RuntimeException(ex);
            }

            return new ResultSetIterator<>(ps, rs -> {
                return resultSetToCell(rs);
            });
        };
    }

    @Override
    public Iterable<ExcelTextStyle> getTextStyleListIter(ExcelCell cell) {
        return () -> {
            String query = getQuery("/de/dfki/sds/hephaistos/storage/excel/SelectTextStyle.sql");
            PreparedStatement ps;
            try {
                ps = connection.prepareStatement(query);
                ps.setInt(1, cell.getId());
            } catch (SQLException ex) {
                throw new RuntimeException(ex);
            }

            return new ResultSetIterator<>(ps, rs -> {
                return resultSetToTextStyle(rs);
            });
        };
    }
    
    @Override
    public ExcelSheet getSheet(int id) {
        return SQLiteUtility.supply(connection, (c) -> {
            String query = getQuery("/de/dfki/sds/hephaistos/storage/excel/SelectSheetWithId.sql");
            PreparedStatement ps = c.prepareStatement(query);
            ps.setInt(1, id);
            ResultSet rs = ps.executeQuery();
            if(rs.next()) {
                ExcelSheet sheet = resultSetToSheet(rs);
                ps.close();
                return sheet;
            }
            return null;
        });
    }

    @Override
    public ExcelCell getCell(int id) {
        return SQLiteUtility.supply(connection, (c) -> {
            String query = getQuery("/de/dfki/sds/hephaistos/storage/excel/SelectCellWithId.sql");
            PreparedStatement ps = c.prepareStatement(query);
            ps.setInt(1, id);
            ResultSet rs = ps.executeQuery();
            if(rs.next()) {
                ExcelCell cell = resultSetToCell(rs);
                ps.close();
                return cell;
            }
            return null;
        });
    }

    @Override
    public void clear() {
        SQLiteUtility.execute(connection, getQuery("/de/dfki/sds/hephaistos/storage/excel/ClearSheet.sql"));
        SQLiteUtility.execute(connection, getQuery("/de/dfki/sds/hephaistos/storage/excel/ClearCell.sql"));
        SQLiteUtility.execute(connection, getQuery("/de/dfki/sds/hephaistos/storage/excel/ClearTextStyle.sql"));
    }

    @Override
    public void remove() {
        SQLiteUtility.execute(connection, getQuery("/de/dfki/sds/hephaistos/storage/excel/DropSheet.sql"));
        SQLiteUtility.execute(connection, getQuery("/de/dfki/sds/hephaistos/storage/excel/DropCell.sql"));
        SQLiteUtility.execute(connection, getQuery("/de/dfki/sds/hephaistos/storage/excel/DropTextStyle.sql"));
    }

    @Override
    public ExcelStorageSummary summary() {
        return new ExcelStorageSummary(this, metaData.getSummaryCache());
    }

    /*
    @Override
    public List<ExcelSheet> search(String keywords) {
        throw new RuntimeException("not implemented yet");
    }

    @Override
    public ResultSet query(String query) {
        throw new RuntimeException("not implemented yet");
    }

    @Override
    public void execute(String query) {
        SQLiteUtility.execute(connection, SQLiteUtility.replaceTablename(query, tablename));
    }
    */
    
    @Override
    public long size() {
        String query = getQuery("/de/dfki/sds/hephaistos/storage/excel/CountSheet.sql");
        return SQLiteUtility.supply(connection, c -> {
            PreparedStatement stmt = c.prepareStatement(query);
            ResultSet rs = stmt.executeQuery();
            long size = rs.getLong(1);
            stmt.close();
            return size;
        });
    }

    @Override
    public void close() {
        //ignore
    }

    private String color2string(Color color) {
        if (color == null) {
            return null;
        }

        return ("" + color.getRed() + ","
                + color.getGreen() + ","
                + color.getBlue() + ","
                + color.getAlpha());
    }

    private Color string2color(String color) {
        if (color == null) {
            return null;
        }

        String[] colorStr = color.split(",");
        return new Color(
                Integer.parseInt(colorStr[0]),
                Integer.parseInt(colorStr[1]),
                Integer.parseInt(colorStr[2]),
                Integer.parseInt(colorStr[3])
        );
    }

    @Override
    public List<ExcelCell> getRow(ExcelSheet sheet, int equalRow, int greaterEqualColumn) {
        
        List<ExcelCell> l = new ArrayList<>();
        try {

            String query = getQuery("/de/dfki/sds/hephaistos/storage/excel/SelectRow.sql");
            PreparedStatement ps;
            try {
                ps = connection.prepareStatement(query);
                ps.setInt(1, sheet.getId());
                ps.setInt(2, equalRow);
                ps.setInt(3, greaterEqualColumn);
            } catch (SQLException ex) {
                throw new RuntimeException(ex);
            }

            ResultSet rs = ps.executeQuery();
            while(rs.next()) {
                l.add(resultSetToCell(rs));
            }

            ps.close();

        } catch (SQLException ex) {
            throw new RuntimeException(ex);
        }
        
        return l;
    }

    @Override
    public List<String> listDeepLinks() {
        List<String> result = new ArrayList<>();
        
        //sheets
        String sheetQuery = getQuery("/de/dfki/sds/hephaistos/storage/excel/SelectSheetId.sql");
        SQLiteUtility.run(connection, (c) -> {
            PreparedStatement ps = c.prepareStatement(sheetQuery);
            ResultSet rs = ps.executeQuery();
            while(rs.next()) {
                result.add(StorageManager.DEEPLINK + "/" + getId() + "/sheet/" + rs.getInt(1));
            }
            ps.close();
        });
        
        
        //cells
        String cellQuery = getQuery("/de/dfki/sds/hephaistos/storage/excel/SelectCellId.sql");
        SQLiteUtility.run(connection, (c) -> {
            PreparedStatement ps = c.prepareStatement(cellQuery);
            ResultSet rs = ps.executeQuery();
            while(rs.next()) {
                result.add(StorageManager.DEEPLINK + "/" + getId() + "/cell/" + rs.getInt(1));
            }
            ps.close();
        });
        
        return result;
    }

}
