package de.dfki.sds.hephaistos.storage.rdf;

import de.dfki.sds.hephaistos.deeplink.DLResource;
import de.dfki.sds.hephaistos.deeplink.DeepLink;
import de.dfki.sds.hephaistos.deeplink.FIChild;
import de.dfki.sds.hephaistos.deeplink.FIToRDF;
import de.dfki.sds.hephaistos.deeplink.Ontology;
import de.dfki.sds.hephaistos.deeplink.RMap;
import de.dfki.sds.hephaistos.storage.InternalStorageMetaData;
import de.dfki.sds.hephaistos.storage.RDFDatasetStorage;
import de.dfki.sds.hephaistos.storage.StorageSummary;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import org.apache.commons.io.FileUtils;
import org.apache.jena.query.Dataset;
import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.rdf.model.RDFNode;
import org.apache.jena.rdf.model.Resource;
import org.apache.jena.rdf.model.ResourceFactory;
import org.apache.jena.rdf.model.SimpleSelector;
import org.apache.jena.system.Txn;
import org.apache.jena.tdb2.TDB2Factory;
import org.apache.jena.vocabulary.RDF;
import org.apache.jena.vocabulary.RDFS;

/**
 * A TDB2 implementation of RDFDatasetStorage.
 * @author Markus Schr&ouml;der
 */
public class RDFDatasetStorageTDB2 extends RDFDatasetStorage implements FIToRDF, FIChild {

    private Dataset dataset;
    private String location;
    
    public RDFDatasetStorageTDB2(InternalStorageMetaData metaData, String location) {
        super(metaData);
        dataset = TDB2Factory.connectDataset(location);
        
        //dataset.addNamedModel(location, model)
        //dataset.containsNamedModel(location)
        //dataset.removeNamedModel(location)
    }
    
    @Override
    public void clear() {
        Txn.executeWrite(dataset, () -> {
            for(String name : getGraphNames()) {
                dataset.removeNamedModel(name);
            }
        });
    }

    @Override
    public void remove() {
        close();
        FileUtils.deleteQuietly(new File(location));
    }

    @Override
    public StorageSummary summary() {
        return null;
    }

    /*
    @Override
    public List<RDFStorageItem> search(String keywords) {
        throw new RuntimeException("not implemented yet");
    }

    @Override
    public ResultSet query(String query) {
        return Txn.calculateRead(dataset, () -> {
            QueryExecution qe = QueryExecutionFactory.create(query, dataset);
            return new ResultSetMem(qe.execSelect());
        });
    }

    @Override
    public void execute(String query) {
        Txn.executeWrite(dataset, () -> {
            UpdateRequest updateRequest = UpdateFactory.create(query);
            UpdateProcessor processor = UpdateExecutionFactory.create(updateRequest, dataset);
            processor.execute();
        });
    }
    */
    
    @Override
    public long size() {
        return getGraphNames().size();
    }

    @Override
    public void close() {
        dataset.close();
    }

    @Override
    public List<String> getGraphNames() {
        return Txn.calculateRead(dataset, () -> {
            List<String> result = new ArrayList<>();
            dataset.listNames().forEachRemaining(result::add);
            return result;
        });
    }

    @Override
    public void addGraph(String uri) {
        Txn.executeWrite(dataset, () -> {
            Model m = ModelFactory.createDefaultModel();
            m.add(ResourceFactory.createResource(uri), RDF.type, RDFS.Container);
            dataset.addNamedModel(uri, m);
        });
    }

    @Override
    public void removeGraph(String uri) {
        Txn.executeWrite(dataset, () -> {
            dataset.removeNamedModel(uri);
        });
    }

    @Override
    public boolean containsGraph(String uri) {
        return Txn.calculateWrite(dataset, () -> {
            return dataset.containsNamedModel(uri);
        });
    }

    @Override
    public void toRDF(Resource resource, Model model, DeepLink deepLink) {
        model.add(resource, RDF.type, Ontology.resource("Dataset"));
        for(String graphUri : getGraphNames()) {
            model.add(resource, Ontology.property("hasGraph"), ResourceFactory.createResource(graphUri));
        }
    }

    @Override
    public DLResource child() {
        return new RMap();
    }

    @Override
    public DLResource child(String name) {
        //String uri = Server.HOST + "/" + Project.PROJECT_SEG + "/" + "storageManager" + "/" + StorageManager.KG_RES + "/" + name;
        //Resource resource = ResourceFactory.createResource(uri);
        return new VirtualResource();
    }

    @Override
    public List<String> listDeepLinks() {
        throw new RuntimeException("not implemented: use listDeepLinks per graph");
    }

    @Override
    public Dataset getDataset() {
        return dataset;
    }
    
    
    private class VirtualResource implements DLResource, FIToRDF {
        
        @Override
        public void toRDF(Resource resource, Model model, DeepLink deepLink) {
            Txn.executeRead(dataset, () -> {
                Model submodel = dataset.getUnionModel().query(new SimpleSelector(resource, null, (RDFNode) null));
                model.add(submodel);
            });
        }
        
    }
    
}
