package de.dfki.sds.hephaistos.storage.rdf;

import de.dfki.sds.hephaistos.storage.InternalStorageMetaData;
import de.dfki.sds.hephaistos.storage.RDFDatasetStorage;
import de.dfki.sds.hephaistos.storage.RDFGraphStorage;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.function.Supplier;
import org.apache.jena.query.Dataset;
import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.NodeIterator;
import org.apache.jena.rdf.model.RDFNode;
import org.apache.jena.rdf.model.ResIterator;
import org.apache.jena.rdf.model.Resource;
import org.apache.jena.rdf.model.ResourceFactory;
import org.apache.jena.rdf.model.Statement;
import org.apache.jena.system.Txn;
import org.apache.jena.tdb2.TDB2Factory;
import org.apache.jena.vocabulary.DCAT;
import org.apache.jena.vocabulary.RDFS;

/**
 * An RDF graph that is backed up by TDB2 on disk.
 * @author Markus Schr&ouml;der
 */
public class RDFGraphStorageTDB2 extends RDFGraphStorage {
    
    private Dataset dataset;
    private Model model;
    
    public RDFGraphStorageTDB2(InternalStorageMetaData metaData, String location) {
        super(metaData);
        dataset = TDB2Factory.connectDataset(location);
        model = dataset.getNamedModel(getId());
    }
    
    private Resource getGraphResource() {
        return ResourceFactory.createResource(getId());
    }
    
    @Override
    public Model getModel() {
        return model;
    }
    
    @Override
    public void executeRead(Runnable r) {
        Txn.executeRead(dataset, r);
    }

    @Override
    public void executeWrite(Runnable r) {
        Txn.executeWrite(dataset, r);
    }

    @Override
    public <T> T calculateRead(Supplier<T> s) {
        return Txn.calculateRead(dataset, s);
    }

    @Override
    public <T> T calculateWrite(Supplier<T> s) {
        return Txn.calculateWrite(dataset, s);
    }
    
    
    @Override
    public void clear() {
        Txn.executeWrite(dataset, () -> {
            //both works
            //dataset.removeNamedModel(getGraphUri());
            model.removeAll();
        });
    }

    @Override
    public void remove() {
        clear();
    }
    
    
    @Override
    public RDFGraphStorageSummary summary() {
        return new RDFGraphStorageSummary(this, metaData.getSummaryCache());
    }

    /*
    @Override
    public List<RDFStorageItem> search(String keywords) {
        throw new RuntimeException("not implemented yet");
    }

    @Override
    public ResultSet query(String query) {
        return Txn.calculateRead(dataset, () -> {
            QueryExecution qe = QueryExecutionFactory.create(query, model); //or dataset?
            return new ResultSetMem(qe.execSelect());
        });
    }

    @Override
    public void execute(String query) {
        Txn.executeWrite(dataset, () -> {
            UpdateRequest updateRequest = UpdateFactory.create(query);
            UpdateProcessor processor = UpdateExecutionFactory.create(updateRequest, dataset);
            processor.execute();
        });
    }
    */
    
    @Override
    public void close() {
        dataset.close();
    }
    
    @Override
    public long size() {
        return Txn.calculateRead(dataset, () -> {
            return model.size();
        });
    }

    @Override
    public void setLabel(String label) {
        executeWrite(() -> {
            model.removeAll(getGraphResource(), RDFS.label, null);
            model.add(getGraphResource(), RDFS.label, label);
        });
    }

    @Override
    public String getLabel() {
        return calculateRead(() -> {
            Statement stmt = model.getProperty(getGraphResource(), RDFS.label);
            if(stmt == null)
                return "";
            return stmt.getObject().asLiteral().getLexicalForm();
        });
    }

    @Override
    public void setPrefix(String prefix) {
        executeWrite(() -> {
            model.removeAll(getGraphResource(), DCAT.keyword, null);
            model.add(getGraphResource(), DCAT.keyword, prefix);
        });
    }

    @Override
    public String getPrefix() {
        return calculateRead(() -> {
            Statement stmt = model.getProperty(getGraphResource(), DCAT.keyword);
            if(stmt == null)
                return "";
            return stmt.getObject().asLiteral().getLexicalForm();
        });
    }

    @Override
    public List<String> listDeepLinks() {
        Set<String> result = new HashSet<>();
        
        executeRead(() -> {
            iter2set(getModel().listSubjects(), result);
            iter2set(getModel().listObjects(), result);
        });
        
        return new ArrayList<>(result);
    }
    
    private void iter2set(ResIterator iter, Set<String> result) {
        while(iter.hasNext()) {
            String uri = iter.next().getURI();
            if(uri.startsWith(RDFDatasetStorage.DEEPLINK)) {
                result.add(uri);
            }
        }
    }
    
    private void iter2set(NodeIterator iter, Set<String> result) {
        while(iter.hasNext()) {
            RDFNode node = iter.next();
            if(node.isResource()) {
                String uri = node.asResource().getURI();
                if(uri.startsWith(RDFDatasetStorage.DEEPLINK)) {
                    result.add(uri);
                }
            }
        }
    }


    

    
}
