package de.dfki.sds.hephaistos.storage.tagfile;

import de.dfki.sds.hephaistos.storage.BranchLeafStorageSqlite;
import de.dfki.sds.hephaistos.storage.BranchLeafStorageSummary;
import de.dfki.sds.hephaistos.storage.InternalStorageMetaData;
import de.dfki.sds.hephaistos.storage.StorageManager;
import de.dfki.sds.hephaistos.storage.TypedName;
import de.dfki.sds.hephaistos.storage.file.*;
import java.nio.file.attribute.FileTime;
import java.sql.Connection;
import java.util.Arrays;
import java.util.List;

/**
 * A file store where the folders are tags based on {@link BranchLeafStorageSqlite}.
 * @author Markus Schr&ouml;der
 */
public class TagFileStorage extends BranchLeafStorageSqlite<Tag, FileInfo, BranchLeafStorageSummary> {

    public TagFileStorage(InternalStorageMetaData metaData, Connection connection) {
        super(metaData, connection);
    }
    
    @Override
    protected List<TypedName> getBranchSchema() {
        return Arrays.asList(
                new TypedName("name", String.class)
        );
    }

    @Override
    protected List<TypedName> getLeafSchema() {
        return getSchema();
    }

    private  List<TypedName> getSchema() {
        return Arrays.asList(
                new TypedName("name", String.class),
                new TypedName("path", String.class),
                new TypedName("lastModifiedTime", Integer.class),
                new TypedName("lastAccessTime", Integer.class),
                new TypedName("creationTime", Integer.class),
                new TypedName("size", Integer.class),
                new TypedName("symbolicLink", Boolean.class)
        );
    }
    
    private Object[] getInsertParams(FileInfo file) {
        return new Object[] { 
            file.getName(), 
            file.getPath(), 
            file.getLastModifiedTimeMillis(),
            file.getLastAccessTimeMillis(), 
            file.getCreationTimeMillis(), 
            file.getSize(), 
            file.isSymbolicLink()
        };
    }
    
    @Override
    protected Object[] getBranchInsertParams(Tag branch) {
        return new Object[] { branch.getName() };
    }

    @Override
    protected Object[] getLeafInsertParams(FileInfo leaf) {
        return getInsertParams(leaf);
    }

    @Override
    protected Class<Tag> getBranchClass() {
        return Tag.class;
    }

    @Override
    protected Class<FileInfo> getLeafClass() {
        return FileInfo.class;
    }

    @Override
    protected Tag getBranchFromRow(Row rs) {
        Tag tag = new Tag();
        tag.setName(rs.getString(1));
        return tag;
    }

    @Override
    protected FileInfo getLeafFromRow(Row rs) {
        return getFromRow(rs);
    }

    private FileInfo getFromRow(Row rs) {
        FileInfo fi = new FileInfo();
        fi.setName(rs.getString(1));
        fi.setPath(rs.getString(2));
        fi.setLastModifiedTime(FileTime.fromMillis(rs.getLong(3)));
        fi.setLastAccessTime(FileTime.fromMillis(rs.getLong(4)));
        fi.setCreationTime(FileTime.fromMillis(rs.getLong(5)));
        fi.setSize(rs.getLong(6));
        fi.setSymbolicLink(rs.getBoolean(7));
        return fi;
    }
    
    @Override
    protected MetaData getBranchMetaData(Tag branch) {
        return new MetaData(branch.getId(), branch.getParent(), branch.getSort());
    }

    @Override
    protected MetaData getLeafMetaData(FileInfo leaf) {
        return new MetaData(leaf.getId(), leaf.getParent(), leaf.getSort());
    }

    @Override
    protected void setBranchMetaData(Tag branch, MetaData metaData) {
        branch.setDeepLink(StorageManager.DEEPLINK + "/" + getId() + "/" + metaData.getId());
        branch.setId(metaData.getId());
        branch.setParent(metaData.getParent());
        branch.setSort(metaData.getSort());
        branch.setType(metaData.getType());
    }

    @Override
    protected void setLeafMetaData(FileInfo leaf, MetaData metaData) {
        leaf.setDeepLink(StorageManager.DEEPLINK + "/" + getId() + "/" + metaData.getId());
        leaf.setId(metaData.getId());
        leaf.setParent(metaData.getParent());
        leaf.setSort(metaData.getSort());
        leaf.setType(metaData.getType());
    }

    @Override
    public BranchLeafStorageSummary summary() {
        return new BranchLeafStorageSummary(this, metaData.getSummaryCache());
    }
    
}
