package de.dfki.sds.hephaistos.storage.term;

import de.dfki.sds.hephaistos.storage.InternalStorageMetaData;
import de.dfki.sds.hephaistos.storage.ResultSetIterator;
import de.dfki.sds.hephaistos.storage.SQLiteUtility;
import de.dfki.sds.hephaistos.storage.StorageManager;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.StringJoiner;
import org.apache.commons.io.IOUtils;

/**
 *
 * @author Markus Schr&ouml;der
 */
public class TermMetricsStorageSqlite extends TermMetricsStorage {

    private Connection connection;
    private String tablename;
    
    public TermMetricsStorageSqlite(InternalStorageMetaData metaData, Connection connection) {
        super(metaData);
        this.connection = connection;
        this.tablename = metaData.getId();
        createTables();
    }
    
    private String getQuery(String path) {
        try {
            String query = IOUtils.toString(getClass().getResourceAsStream(path), StandardCharsets.UTF_8);
            query = query.replaceAll("\\$\\{tablename\\}", tablename);
            return query;
        } catch (IOException ex) {
            throw new RuntimeException(ex);
        }
    }
    
    private void createTables() {
        SQLiteUtility.execute(connection, getQuery("/de/dfki/sds/hephaistos/storage/term/CreateTable.sql"));
    }

    @Override
    public void addBulk(Collection<TermMetrics> items) {
        String query = getQuery("/de/dfki/sds/hephaistos/storage/term/Insert.sql");

        SQLiteUtility.run(connection, c -> {
            c.setAutoCommit(false);

            PreparedStatement ps = c.prepareStatement(query);

            for (TermMetrics term : items) {

                ps.setInt(1, term.getId());
                ps.setString(2, term.getTerm());
                
                /*
                ?, -- "ttf" REAL,
                ?, -- "atf" REAL,
                ?, -- "idf" REAL,
                ?, -- "atf_idf" REAL,
                ?, -- "ttf_idf" REAL,
                ?, -- "ridf" REAL,
                ?, -- "cvalue" REAL,
                ?, -- "rake_frequency" REAL,
                ?, -- "rake_degree" REAL,
                ? -- "rake_degree_frequency" REAL
                */
                
                ps.setDouble(3, term.getTTF());
                ps.setDouble(4, term.getATF());
                ps.setDouble(5, term.getIDF());
                ps.setDouble(6, term.getATF_IDF());
                ps.setDouble(7, term.getTTF_IDF());
                ps.setDouble(8, term.getRIDF());
                ps.setDouble(9, term.getCValue());
                ps.setDouble(10, term.getRAKE_Frequency());
                ps.setDouble(11, term.getRAKE_Degree());
                ps.setDouble(12, term.getRAKE_Degree_Frequency());

                ps.addBatch();
            }

            ps.executeBatch();
            c.commit();
            ps.close();
            c.setAutoCommit(true);
        });
    }

    @Override
    public void removeBulk(Collection<TermMetrics> items) {
        if (items.isEmpty()) {
            return;
        }

        StringJoiner sj = new StringJoiner(",");
        for (TermMetrics term : items) {
            sj.add(String.valueOf(term.getId()));
        }
        String ids = sj.toString();

        String query = getQuery("/de/dfki/sds/hephaistos/storage/term/Delete.sql");
        query = query.replace("${ids}", ids);
        SQLiteUtility.execute(connection, query);
    }

    @Override
    public Iterable<TermMetrics> getListIter() {
        return () -> {
            String query = getQuery("/de/dfki/sds/hephaistos/storage/term/Select.sql");
            PreparedStatement ps;
            try {
                ps = connection.prepareStatement(query);
            } catch (SQLException ex) {
                throw new RuntimeException(ex);
            }

            return new ResultSetIterator<>(ps, rs -> {
                return resultSetToTerm(rs);
            });
        };
    }

    private TermMetrics resultSetToTerm(ResultSet rs) {
        TermMetrics term = new TermMetrics();
        try {
            term.setDeepLink(StorageManager.DEEPLINK + "/" + getId() + "/" + rs.getInt(1));
            
            term.setId(rs.getInt(1));

            term.setTerm(rs.getString(2));

            /*
            ?, -- "ttf" REAL,
            ?, -- "atf" REAL,
            ?, -- "idf" REAL,
            ?, -- "atf_idf" REAL,
            ?, -- "ttf_idf" REAL,
            ?, -- "ridf" REAL,
            ?, -- "cvalue" REAL,
            ?, -- "rake_frequency" REAL,
            ?, -- "rake_degree" REAL,
            ? -- "rake_degree_frequency" REAL
            */
            
            term.setTTF(rs.getDouble(3));
            term.setATF(rs.getDouble(4));
            term.setIDF(rs.getDouble(5));
            term.setATF_IDF(rs.getDouble(6));
            term.setTTF_IDF(rs.getDouble(7));
            term.setRIDF(rs.getDouble(8));
            term.setCValue(rs.getDouble(9));
            term.setRAKE_Frequency(rs.getDouble(10));
            term.setRAKE_Degree(rs.getDouble(11));
            term.setRAKE_Degree_Frequency(rs.getDouble(12));
            
        } catch (SQLException ex) {
            throw new RuntimeException(ex);
        }
        return term;
    }
    
    @Override
    public void clear() {
        SQLiteUtility.execute(connection, getQuery("/de/dfki/sds/hephaistos/storage/term/Clear.sql"));
    }

    @Override
    public void remove() {
        SQLiteUtility.execute(connection, getQuery("/de/dfki/sds/hephaistos/storage/term/Drop.sql"));
    }

    @Override
    public TermMetricsStorageSummary summary() {
        return null;
    }

    /*
    @Override
    public List<TermMetrics> search(String keywords) {
        throw new RuntimeException("not implemented yet");
    }

    @Override
    public ResultSet query(String query) {
        throw new RuntimeException("not implemented yet");
    }

    @Override
    public void execute(String query) {
        SQLiteUtility.execute(connection, SQLiteUtility.replaceTablename(query, tablename));
    }
    */
    
    @Override
    public long size() {
        String query = getQuery("/de/dfki/sds/hephaistos/storage/term/Count.sql");
        return SQLiteUtility.supply(connection, c -> {
            PreparedStatement ps = c.prepareStatement(query);
            ResultSet rs = ps.executeQuery();
            long size = rs.getLong(1);
            ps.close();
            return size;
        });
    }

    @Override
    public void close() {
        //ignore
    }

    @Override
    public List<String> listDeepLinks() {
        String query = getQuery("/de/dfki/sds/hephaistos/storage/term/SelectId.sql");
        
        List<String> result = new ArrayList<>();
        
        SQLiteUtility.run(connection, (c) -> {
            PreparedStatement ps = c.prepareStatement(query);
            ResultSet rs = ps.executeQuery();
            while(rs.next()) {
                result.add(StorageManager.DEEPLINK + "/" + getId() + "/" + rs.getInt(1));
            }
            ps.close();
        });
        
        return result;
    }
    
}
