package de.dfki.sds.mschroeder.commons.lang;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Utility for calculating combinatorics.
 * @author Markus Schr&ouml;der
 */
public class CombinatoricsUtility {
    
    public static <T> Set<Set<T>> cartesianProductSet(Set<T>... sets) {
    if (sets.length < 2)
        throw new IllegalArgumentException(
                "Can't have a product of fewer than two sets (got " +
                sets.length + ")");

    return _cartesianProductSet(0, sets);
}

    private static <T> Set<Set<T>> _cartesianProductSet(int index, Set<T>... sets) {
        Set<Set<T>> ret = new HashSet<>();
        if (index == sets.length) {
            ret.add(new HashSet<>());
        } else {
            for (T obj : sets[index]) {
                for (Set<T> set : _cartesianProductSet(index+1, sets)) {
                    set.add(obj);
                    ret.add(set);
                }
            }
        }
        return ret;
    }
    
    public static <T> List<List<T>> cartesianProductList(List<List<T>> lists) {
    if (lists.size() < 2)
        throw new IllegalArgumentException(
                "Can't have a product of fewer than two lists (got " +
                lists.size() + ")");

    return _cartesianProductList(0, lists);
}

    private static <T> List<List<T>> _cartesianProductList(int index, List<List<T>> lists) {
        List<List<T>> ret = new LinkedList<>();
        if (index == lists.size()) {
            ret.add(new LinkedList<>());
        } else {
            for (T obj : lists.get(index)) {
                for (List<T> set : _cartesianProductList(index+1, lists)) {
                    set.add(obj);
                    ret.add(set);
                }
            }
        }
        return ret;
    }
    
    
    //https://www.kite.com/python/answers/how-to-get-all-unique-combinations-of-two-lists-in-python
    public static <T> List<Map<T, T>> possibleMatches(List<T> a, List<T> b) {
        
        List<Map<T, T>> result = new ArrayList<>();

        boolean swap = false;

        //swap
        if (b.size() > a.size()) {
            List<T> tmp = a;
            a = b;
            b = tmp;

            swap = true;
        }
        
        List<List<T>> permutations = choose(a, b.size());
        
        for(List<T> permutation : permutations) {

            Map<T, T> map = new HashMap<>();

            for (int i = 0; i < permutation.size(); i++) {

                T t1 = permutation.get(i);
                T t2 = b.get(i);

                if (swap) {
                    map.put(t2, t1);
                } else {
                    map.put(t1, t2);
                }
            }

            result.add(map);
        }

        return result;
    }
    
    
    //from https://stackoverflow.com/questions/44949030/print-all-possible-permutations-of-r-elements-in-a-given-integer-array-of-size-n
    public static <T> List<List<T>> choose(List<T> a, int r) {
        List<List<T>> allPermutations = new ArrayList<>();
        enumerate(a, a.size(), r, allPermutations);
        return allPermutations;
    }

    private static <T> void enumerate(List<T> a, int n, int k, List<List<T>> allPermutations) {
        if (k == 0) {
            ArrayList<T> singlePermutation = new ArrayList<>();
            for (int i = n; i < a.size(); i++){
                singlePermutation.add(a.get(i));
            }
            allPermutations.add(singlePermutation);
            return;
        }

        for (int i = 0; i < n; i++) {
            swap(a, i, n-1);
            enumerate(a, n-1, k-1, allPermutations);
            swap(a, i, n-1);
        }
    }  

    private static <T> void swap(List<T> a, int i, int j) {
        T temp = a.get(i);
        a.set(i, a.get(j));
        a.set(j, temp);
    }
    
}
