package de.dfki.sds.mschroeder.commons.lang;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.FileSystem;
import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.function.Consumer;
import java.util.stream.Stream;
import java.util.zip.GZIPInputStream;

/**
 *
 * @author Markus Schr&ouml;der
 */
public class FileUtility {

    public static String relativePathToWorkingDir(File file) {
        try {
            Path pathAbsolute = file.getCanonicalFile().toPath();
            Path pathBase = new File(".").getCanonicalFile().toPath();
            Path pathRelative = pathBase.relativize(pathAbsolute);
            return pathRelative.toString();
        } catch (IOException ex) {
            throw new RuntimeException(ex);
        }
    }

    public static File file(File start, String... parts) {
        File cur = start;
        for (int i = 0; i < parts.length; i++) {
            cur = new File(cur, parts[i]);
        }
        return cur;
    }

    public static void crawl(File root, Integer maxDepth, Consumer<File> fileConsumer) {
        crawl(Arrays.asList(root), maxDepth, null, false, fileConsumer);
    }

    public static void crawl(List<File> roots, Integer maxDepth, FileFilter filefilter, boolean allowHidden, Consumer<File> fileConsumer) {
        for (File root : roots) {
            File cur = root;

            fileConsumer.accept(cur);

            Queue<FileDepth> q = new LinkedList<>();
            q.add(new FileDepth(cur, null, 0));

            while (!q.isEmpty()) {
                FileDepth fd = q.poll();

                if (maxDepth != null && fd.depth >= maxDepth) {
                    break;
                }

                //children
                File[] children;
                if (filefilter != null) {
                    children = fd.file.listFiles(filefilter);
                } else {
                    children = fd.file.listFiles();
                }

                //process child and recursive depth+1
                if (children != null) {
                    for (File child : children) {

                        if (!allowHidden && child.isHidden()) {
                            continue;
                        }

                        fileConsumer.accept(child);
                        q.add(new FileDepth(child, fd.file, fd.depth + 1));
                    }
                }
            }
        }
    }

    public static Iterable<File> crawl(File root, Integer maxDepth, FileFilter filefilter, boolean allowHidden) {
        return () -> {
            return new Iterator<File>() {

                Queue<FileDepth> q = new LinkedList<>(Arrays.asList(new FileDepth(root, null, 0)));

                @Override
                public boolean hasNext() {
                    return !q.isEmpty();
                }

                @Override
                public File next() {
                    FileDepth fd = q.poll();
                    step(fd);
                    return fd.file;
                }

                //returns continue = {true, false}
                private void step(FileDepth fd) {
                    if (maxDepth != null && fd.depth >= maxDepth) {
                        return;
                    }

                    //children
                    File[] children;
                    if (filefilter != null) {
                        children = fd.file.listFiles(filefilter);
                    } else {
                        children = fd.file.listFiles();
                    }

                    //process child and recursive depth+1
                    if (children != null) {
                        for (File child : children) {

                            if (!allowHidden && child.isHidden()) {
                                continue;
                            }

                            q.add(new FileDepth(child, fd.file, fd.depth + 1));
                        }
                    }
                }
            };
        };
    }

    public static List<File> crawlList(File root, Integer maxDepth, FileFilter filefilter, boolean allowHidden) {
        List<File> f = new ArrayList<>();
        crawl(root, maxDepth, filefilter, allowHidden).forEach(f::add);
        return f;
    }

    //public becaues also used by FilesystemLoader
    public static class FileDepth {

        public File file;
        public File parent;
        public int depth;

        public FileDepth(File file, File parent, int depth) {
            this.file = file;
            this.parent = parent;
            this.depth = depth;
        }
    }

    public static List<String> crawl(String resourceRoot) {

        List<String> paths = new LinkedList<>();

        try {
            URI uri = Files.class.getResource(resourceRoot).toURI();
            Path myPath;
            if (uri.getScheme().equals("jar")) {
                FileSystem fileSystem = FileSystems.newFileSystem(uri, Collections.<String, Object>emptyMap());
                myPath = fileSystem.getPath(resourceRoot);
            } else {
                myPath = Paths.get(uri);
            }
            Stream<Path> walk = Files.walk(myPath);
            for (Iterator<Path> it = walk.iterator(); it.hasNext();) {
                Path next = it.next();
                paths.add(resourceRoot + next.toString().substring(myPath.toString().length(), next.toString().length()));
            }

        } catch (URISyntaxException | IOException ex) {
            throw new RuntimeException(ex);
        }

        return paths;
    }

    // https://stackoverflow.com/questions/453018/number-of-lines-in-a-file-in-java
    public static int countLines(File file) {
        if (file.isDirectory()) {
            return 0;
        }

        InputStream is = null;
        try {
            is = new FileInputStream(file);
            if(file.getName().endsWith("gz")) {
                is = new GZIPInputStream(is);
            }
            is = new BufferedInputStream(is);
            byte[] c = new byte[1024];

            int readChars = is.read(c);
            if (readChars == -1) {
                // bail out if nothing to read
                return 0;
            }

            // make it easy for the optimizer to tune this loop
            int count = 0;
            while (readChars == 1024) {
                for (int i = 0; i < 1024;) {
                    if (c[i++] == '\n') {
                        ++count;
                    }
                }
                readChars = is.read(c);
            }

            // count remaining characters
            while (readChars != -1) {
                //System.out.println(readChars);
                for (int i = 0; i < readChars; ++i) {
                    if (c[i] == '\n') {
                        ++count;
                    }
                }
                readChars = is.read(c);
            }

            return count == 0 ? 1 : count;
        } catch (IOException ex) {
            throw new RuntimeException(ex);
        } finally {
            if (is != null) {
                try {
                    is.close();
                } catch (IOException ex) {
                    throw new RuntimeException(ex);
                }
            }
        }

    }
}
