package de.dfki.sds.mschroeder.commons.lang.data;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringJoiner;
import static java.util.stream.Collectors.toList;

/**
 * Useful for evaluation results.
 * Came from RDF2RDB eval code.
 */
public class LatexTable {
    
    private List<Column> columns;
    private List<Object[]> rows;
    
    public LatexTable() {
        columns = new ArrayList<>();
        rows = new ArrayList<>();
    }
    
    public void addColumn(String abbrev, String alignment) {
        columns.add(new Column(abbrev, alignment));
    }
    
    public void addColumn(String abbrev, String description, String alignment, boolean showInLatexTable) {
        columns.add(new Column(abbrev, description, alignment, showInLatexTable));
    }
    
    public void addRow(Object... data) {
        rows.add(data);
    }
    
    private class Column {
        String abbrev;
        String description;
        String alginment;
        boolean showInLatexTable;

        public Column(String abbrev, String alginment) {
            this.abbrev = abbrev;
            this.alginment = alginment;
        }

        public Column(String abbrev, String description, String alginment, boolean showInLatexTable) {
            this.abbrev = abbrev;
            this.description = description;
            this.alginment = alginment;
            this.showInLatexTable = showInLatexTable;
        }
        
    }
    
    public String getCaption() {
        StringBuilder sb = new StringBuilder();
        sb.append("\\caption{\n");
        
        List<Column> filtered = columns.stream().filter(c -> c.showInLatexTable).collect(toList());
        
        for(int i = 0; i < filtered.size(); i++) {
            Column c = filtered.get(i);
            if(c.showInLatexTable) {
                //sb.append(c.abbrev).append(" -- ").append(c.description);
                sb.append(c.description).append(" (").append(c.abbrev).append(")");
                
                if(i != filtered.size() - 1)
                    sb.append(",\n");
            }
        }
        sb.append("\n}");
        return sb.toString();
    }
    
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        
        StringJoiner alignHeader = new StringJoiner("|", "|", "|");
        List<String> colAbbrevs = new ArrayList<>();
        for(Column col : columns) {
            colAbbrevs.add(col.abbrev);
            
            if(!col.showInLatexTable)
                continue;
            
            alignHeader.add(col.alginment);
        }

        Map<Integer, Integer> col2maxlen = new HashMap<>();
        List<Object[]> rowsTmpList = new ArrayList<>(rows);
        rowsTmpList.add(colAbbrevs.toArray());
        for(Object[] row : rowsTmpList) {
            for(int i = 0; i < columns.size(); i++) {
                Object obj = row[i];
                String str = obj.toString();
                int maxlen = col2maxlen.computeIfAbsent(i, c -> 0);
                
                if(str.length() > maxlen) {
                    col2maxlen.put(i, str.length());
                }
            }
        }
        
        sb.append("\\begin{tabular}{"+alignHeader.toString()+"}\n");
        sb.append("\\hline\n");
        StringJoiner header1 = new StringJoiner(" & ");
        for(int i = 0; i < columns.size(); i++) {
            Column col = columns.get(i);
            
            if(!col.showInLatexTable)
                continue;
            
            int maxlen = col2maxlen.get(i);
            int curlen = col.abbrev.length();
            int difflen = maxlen - curlen;

            String indent = new String(new char[difflen]).replace("\0", " ");
            
            header1.add(col.abbrev + indent);
        }
        sb.append(header1.toString() + " \\\\\n");
        sb.append("\\hline\n");
        sb.append("\\hline\n");
        
        for(Object[] row : rows) {
            StringJoiner sj = new StringJoiner(" & ");
            
            for(int i = 0; i < row.length; i++) {
                if(!columns.get(i).showInLatexTable)
                    continue;
                
                int maxlen = col2maxlen.get(i);
                int curlen = row[i].toString().length();
                int difflen = maxlen - curlen;
                
                String indent = new String(new char[difflen]).replace("\0", " ");
                
                sj.add(row[i].toString() + indent);
            }
            
            sb.append(sj.toString() + " \\\\");
            sb.append("\\hline\n");
        }
        
        sb.append("\\end{tabular}\n");
        
        return sb.toString();
    }
    
}
