package de.dfki.sds.mschroeder.commons.lang.math;

import de.dfki.sds.mschroeder.commons.lang.SetUtility;
import java.util.HashSet;
import java.util.Set;

/**
 *
 * @author Markus Schr&ouml;der
 */
public class PrecisionRecall<T> {

    //https://en.wikipedia.org/wiki/Precision_and_recall
    
    private Set<T> relevant;
    private Set<T> retrieved;

    public PrecisionRecall() {
        this.relevant = new HashSet<>();
        this.retrieved = new HashSet<>();
    }
    
    public PrecisionRecall(Set<T> relevant, Set<T> retrieved) {
        this.relevant = relevant;
        this.retrieved = retrieved;
    }

    /**
     * True positive: all hits.
     *
     * @return
     */
    public Set<T> getTP() {
        return new HashSet<>(SetUtility.intersection(relevant, retrieved));
    }

    /**
     * False positive: Type I error: eqv with false alarm. Retrieved documents
     * which was not necessary to find.
     *
     * @return
     */
    public Set<T> getFP() {
        return new HashSet<>(SetUtility.subtract(retrieved, relevant));
    }

    /**
     * False negative: Type II error: eqv with miss. Relevant documents which
     * was not found.
     *
     * @return
     */
    public Set<T> getFN() {
        return new HashSet<>(SetUtility.subtract(relevant, retrieved));
    }

    public Set<T> getRelevant() {
        return relevant;
    }

    public Set<T> getRetrieved() {
        return retrieved;
    }
    
    public Values calculateValues() {
        
        double tp = getTP().size();
        double fp = getFP().size();
        double fn = getFN().size();
        double tn = 0; //in IR we do not have negative examples, or?
        
        double precision = tp / (tp + fp);
        double recall = tp / (tp + fn);
        
        double accuracy = (tp + tn) / (tp + tn + fp + fn);
        
        double fScore = 2 * (precision * recall) / (precision + recall);
        if(Double.isNaN(fScore)) {
            fScore = 0.0;
        }
        
        return new Values(tp, fp, fn, precision, recall, accuracy, fScore);
    }

    public class Values {

        private double tp;
        private double fp;
        private double fn;
        private double precision;
        private double recall;
        private double accuracy;
        private double fScore;

        public Values(double tp, double fp, double fn, double precision, double recall, double accuracy, double fScore) {
            this.tp = tp;
            this.fp = fp;
            this.fn = fn;
            this.precision = precision;
            this.recall = recall;
            this.accuracy = accuracy;
            this.fScore = fScore;
        }
        
        

        public double getTp() {
            return tp;
        }

        public double getFp() {
            return fp;
        }

        public double getFn() {
            return fn;
        }

        public double getPrecision() {
            return precision;
        }

        public double getRecall() {
            return recall;
        }

        public double getAccuracy() {
            return accuracy;
        }

        public double getfScore() {
            return fScore;
        }

        @Override
        public String toString() {
            return "PrecisionRecall.Values{" + "tp=" + tp + ", fp=" + fp + ", fn=" + fn + ", precision=" + precision + ", recall=" + recall + ", accuracy=" + accuracy + ", fScore=" + fScore + '}';
        }
        
    }

    @Override
    public String toString() {
        return calculateValues().toString();
    }
    
}
