package de.dfki.sds.mschroeder.commons.lang.math;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import static java.util.stream.Collectors.toList;

/**
 * A list of scores for an arbitrary object of type T.
 * @author Markus Schr&ouml;der
 * @param <T>
 */
public class Scoring<T> {
    
    private T object;
    private Map<String, Score> scoreMap;
    
    public Scoring(T object) {
        this.object = object;
        this.scoreMap = new HashMap<>();
    }
    
    public Score getScore(String label) {
        return scoreMap.get(label);
    }
    
    public void addScore(String label, String comment, String explaination, double value) {
        scoreMap.put(label, new Score(label,comment, explaination, value));
    }
    
    public void addScore(String label, String comment, String explaination, boolean value) {
        addScore(label, comment, explaination, value ? 1 : 0);
    }
    
    public void increaseScore(String label, String comment, String explaination, double offset, double delta) {
        Score score = getScore(label);
        if(score != null) {
            score.setValue(score.getValue() + delta);
        } else {
            addScore(label, comment, explaination, offset + delta);
        }
    }
    
    public void increaseScore(String label, String comment, String explaination, double offset, boolean bool) {
        increaseScore(label, comment, explaination, offset, bool ? 1 : 0);
    }
    
    public void addScore(Score<T> score) {
        scoreMap.put(score.getLabel(), score);
    }
    
    
    
    public Collection<?> getRow() {
        List<Object> l = new LinkedList<>();
        l.add(object);
        l.addAll(getSortedScores());
        return l;
    }
    
    public Collection<?> getRow(List<String> headers, double missingValue) {
        List<Object> l = new LinkedList<>();
        l.add(object);
        l.addAll(getValues(headers, missingValue));
        return l;
    }
    
    public Collection<?> getRow(Function<T, String> t2string, List<String> headers, double missingValue) {
        List<Object> l = new LinkedList<>();
        l.add(t2string.apply(object));
        l.addAll(getValues(headers, missingValue));
        return l;
    }
    
    public T getObject() {
        return object;
    }

    public List<Score> getScores() {
        return new ArrayList<>(scoreMap.values());
    }
    
    public List<Score> getSortedScores() {
        return scoreMap.values().stream().sorted((o1, o2) -> {
            return o1.getLabel().compareTo(o2.getLabel());
        }).collect(toList());
    }
    
    public List<Double> getValues(List<String> headers, double missingValue) {
        List<Double> d = new LinkedList<>();
        for(String header : headers) {
            Score s = getScore(header);
            
            if(s == null)
                d.add(missingValue);
            else
                d.add(s.getValue());
        }
        return d;
    }
    
    
    
    @Override
    public String toString() {
        return "Scoring{" + "object=" + object + ", scores=" + getScores() + '}';
    }
    
}
