package de.dfki.sds.mschroeder.commons.lang.math;

import de.dfki.sds.mschroeder.commons.lang.StatisticUtility;
import de.dfki.sds.mschroeder.commons.lang.data.Sort;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;
import static java.util.stream.Collectors.toList;

/**
 *
 * @author Markus Schr&ouml;der
 * @param <T>
 */
public class ScoringManager<T> {

    public static final Score ITEM = new Score("item", "the item that is scored", "");

    private List<Scoring<T>> scorings;

    public ScoringManager() {
        this.scorings = new ArrayList<>();
    }

    public ScoringManager(List<Scoring<T>> scorings) {
        this.scorings = scorings;
    }

    
    
    public Scoring<T> addScoring(T object) {
        Scoring<T> scoring = new Scoring<>(object);
        this.scorings.add(scoring);
        return scoring;
    }

    public List<Scoring<T>> addScorings(List<T> objects) {
        for (T t : objects) {
            addScoring(t);
        }
        return scorings;
    }

    public void remove(String scoreLabel, Predicate<Score<T>> filter) {
        scorings.removeIf(scoring -> filter.test(scoring.getScore(scoreLabel)));
    }

    //adds scores for each scoring
    public void score(Consumer<Scoring<T>> scoreFuction) {
        for (Scoring<T> scoring : getScorings()) {
            scoreFuction.accept(scoring);
        }
    }

    //creates a new score from already existing scores
    public void metascore(Function<Double[], Double> calculation, Score metaScoreDesc, String... scoreLabels) {
        score((scoring) -> {

            Score metaScore = metaScoreDesc.copyWithValue(0);

            Double[] values = new Double[scoreLabels.length];
            for (int i = 0; i < scoreLabels.length; i++) {
                //cleanup because one can write "-a*0.3"
                Score s = scoring.getScore(ScoreLabel.cleanup(scoreLabels[i]));

                //if we miss a score we can not calculate the meta score
                if (s == null) {
                    return;
                }

                //all the values
                values[i] = s.getValue();

                //to know where this score comes from
                metaScore.getParents().add(s);
            }

            //set the value of the meta score based on a given calculation
            metaScore.setValue(calculation.apply(values));

            //add meta score to the scoring
            scoring.addScore(metaScore);
        });
    }

    //creates harmonic mean from already existing scores
    public void harmonicMean(Score metaScoreDesc, String... scoreLabels) {
        metascore((values) -> {

            Map<String, Object> key2value = new HashMap<>();
            for (int i = 0; i < scoreLabels.length; i++) {
                key2value.put(scoreLabels[i], values[i]);
            }
            return StatisticUtility.harmonicMeanOf(key2value, scoreLabels);

        }, metaScoreDesc, scoreLabels);
    }

    //creates sum from already existing scores
    public void sum(Score metaScoreDesc, String... scoreLabels) {
        metascore((values) -> {
            double sum = 0;
            for (double d : values) {
                sum += d;
            }
            return sum;
        }, metaScoreDesc, scoreLabels);
    }

    //TODO is only for one label, check max == min
    public void normalize(Score metaScoreDesc, String sourceScoreLabel) {
        double minValue = Double.POSITIVE_INFINITY;
        double maxValue = Double.NEGATIVE_INFINITY;

        for (Scoring<T> scoring : getScorings()) {
            Score score = scoring.getScore(sourceScoreLabel);
            if (score != null) {
                minValue = Math.min(score.getValue(), minValue);
                maxValue = Math.max(score.getValue(), maxValue);
            }
        }

        double min = minValue;
        double max = maxValue;

        metascore((values) -> {
            //https://stats.stackexchange.com/questions/70801/how-to-normalize-data-to-0-1-range
            return (values[0] - min) / (max - min);
        }, metaScoreDesc, sourceScoreLabel);
    }

    public void sort(String scoreLabel, Sort sort) {
        if (sort == Sort.None) {
            return;
        }

        scorings.sort((o1, o2) -> {
            return (sort == Sort.Descending ? -1 : 1) * Double.compare(o1.getScore(scoreLabel).getValue(), o2.getScore(scoreLabel).getValue());
        });
    }

    public void top(int k) {
        scorings = scorings.subList(0, k);
    }

    
    
    public List<String> allSortedScoreLabels() {
        return scorings.stream()
                .flatMap((t) -> {
                    return t.getScores().stream();
                })
                .distinct()
                .sorted()
                .map(score -> score.getLabel())
                .collect(toList());
    }

    public List<String> allSortedScoreComments() {
        return scorings.stream()
                .flatMap((t) -> {
                    return t.getScores().stream();
                })
                .distinct()
                .sorted()
                .map(score -> score.getComment())
                .collect(toList());
    }

    public List<Score> allSortedScores() {
        return scorings.stream()
                .flatMap((t) -> {
                    return t.getScores().stream();
                })
                .distinct()
                .sorted()
                .collect(toList());
    }

    public String toTree(String scoreLabel) {
        return scorings.stream()
                .flatMap((t) -> {
                    return t.getScores().stream();
                })
                .distinct()
                .filter(s -> s.getLabel().equals(scoreLabel))
                .findFirst().get().toStringTree();
    }

    public String toTrees() {
        StringBuilder sb = new StringBuilder();
        List<Score> scores = allSortedScores();
        for (Score s : scores) {
            sb.append(s.toStringTree()).append("\n");
        }
        return sb.toString();
    }

    public int size() {
        return scorings.size();
    }

    public List<Scoring<T>> getScorings() {
        return scorings;
    }

    //TODO make this public available?
    private static class ScoreLabel {

        String label;
        boolean invert;
        double weight;

        static ScoreLabel toScoreLabel(String scoreLabel) {
            ScoreLabel sl = new ScoreLabel();

            sl.invert = false;
            if (scoreLabel.startsWith("-")) {
                scoreLabel = scoreLabel.substring(1);
                sl.invert = true;
            }

            sl.weight = 1.0;
            if (scoreLabel.contains("*")) {
                String[] split = scoreLabel.split("\\*");
                sl.weight = Double.parseDouble(split[1]);
                scoreLabel = split[0];
            }

            sl.label = scoreLabel;

            return sl;
        }

        static String cleanup(String scoreLabel) {
            return toScoreLabel(scoreLabel).label;
        }

    }
    
    //deprecated becuse ascii table, CSV dependency
    /*
    public void printSimple() {
        AsciiTable t = new AsciiTable();
        for(Scoring<T> scoring : scorings) {
            t.addRow(scoring.getRow());
        }
        System.out.println(t.render());
    }
    
    public void print() {
        
        List<String> labels = allSortedScoreLabels();
        List<String> comments = allSortedScoreComments();
        
        List<String> allLabels = new LinkedList<>();
        allLabels.add("item");
        allLabels.addAll(labels);
        
        List<String> allComments = new LinkedList<>();
        allComments.add("the scored item");
        allComments.addAll(comments);
        
        AsciiTable t = new AsciiTable();
        
        //t.addRow(allLabels);
        //t.addRule();
        //t.addRule();
        for(Scoring<T> scoring : scorings) {
            t.addRow(allLabels);
            t.addRow(allComments);
            
            t.addRow(scoring.getRow(labels, Double.NaN));
            
            t.addRule();
        }
        
        System.out.println(scorings.size() + " rows");
        System.out.println(t.render());
    }
     */
 /*
    public void toCSV(Function<T, String> t2string, File csvFile) throws IOException {
        CSVPrinter csv = CSVFormat.DEFAULT.print(new FileWriter(csvFile));
        
        List<Score> scores = allSortedScores();
        
        List<Score> headers = new LinkedList<>();
        headers.add(ITEM);
        headers.addAll(scores);
        
        List<Function<Score, String>> toStringer = Arrays.asList(
                score -> score.getLabel(),
                score -> score.getComment(),
                score -> score.getExplaination()
        );
        
        for(Function<Score, String> toString : toStringer) {
            for(Score score : headers) {
                csv.print(toString.apply(score));
            }
            csv.println();
        }
        
        for(Scoring<T> scoring : scorings) {
            csv.printRecord(scoring.getRow(
                    t2string,
                    scores.stream().map(s -> s.getLabel()).collect(toList()), 
                    Double.NaN)
            );
        }
        
        csv.close();
    }
    
    public JSONObject toJson(Function<T, JSONObject> t2jo, boolean headerAsObjects) {
        List<Score> scores = allSortedScores();
        List<Score> headers = new LinkedList<>();
        headers.add(ITEM);
        headers.addAll(scores);
        
        JSONArray head = new JSONArray();
        if(headerAsObjects) {
            for(Score score : headers) {
                head.put(score.toJson());
            }
        } else {
            List<Function<Score, String>> toStringer = Arrays.asList(
                score -> score.getLabel(),
                score -> score.getComment(),
                score -> score.getExplaination()
            );

            for(Function<Score, String> toString : toStringer) {
                JSON row = JSON.Array();
                for(Score score : headers) { 
                    row.add(toString.apply(score));
                }
                head.put(row.asJSONArray());
            }
        }
        
        List<String> headerLabels = scores.stream().map(s -> s.getLabel()).collect(toList());
        
        JSONArray body = new JSONArray();
        for(Scoring<T> scoring : scorings) {
            body.put(
                    JSON.Object(
                            "item", t2jo.apply(scoring.getObject()),
                            "data", scoring.getRowAsJsonArray(headerLabels)
                    ).asJSONObject()
            );
        }
        
        JSONObject jo = new JSONObject();
        jo.put("head", head);
        jo.put("body", body);
        return jo;
    }
     */
    
}
